#
# Copyright 2004-2013 University of Zagreb.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# This work was supported in part by the Croatian Ministry of Science
# and Technology through the research contract #IP-2003-143.
#

# $Id: editor.tcl 147 2015-03-27 14:37:19Z denis $


#****h* imunes/editor.tcl
# NAME
#  editor.tcl -- file used for defining functions that can be used in
#  edit mode as well as all the functions which change the appearance 
#  of the imunes GUI.
# FUNCTION
#  This module is used for defining all possible actions in imunes
#  edit mode. It is also used for all the GUI related actions.
#****

#****f* editor.tcl/animateCursor
# NAME
#   animateCursor -- animate current cursor
# SYNOPSIS
#   animateCursor
# FUNCTION
#   Animates the cursor to show the status of the running script.
#****
proc animateCursor {} {
    global cursorState
    global clock_seconds

    if { [clock seconds] == $clock_seconds } {
	update
	return
    }
    set clock_seconds [clock seconds]
    if { $cursorState } {
	.panwin.f1.c config -cursor watch
	set cursorState 0
    } else {
	.panwin.f1.c config -cursor pirate
	set cursorState 1
    }
    update
}

#****f* editor.tcl/removeGUILink
# NAME
#   removeGUILink -- remove link from GUI
# SYNOPSIS
#   renoveGUILink $link_id $atomic
# FUNCTION
#   Removes link from GUI. It removes standard links as well as
#   split links and links connecting nodes on different canvases.
# INPUTS
#   * link_id -- the link id
#   * atomic -- defines if the remove was atomic action or a part 
#     of a composed, non-atomic action (relevant for updating log 
#     for undo).
#****
proc removeGUILink { link atomic } {
    global changed

    set nodes [linkPeers $link]
    set node1 [lindex $nodes 0]
    set node2 [lindex $nodes 1]
    if { [nodeType $node1] == "pseudo" } {
	removeLink [getLinkMirror $link]
	removeLink $link
	removeNode [getNodeMirror $node1]
	removeNode $node1
	.panwin.f1.c delete $node1
    } elseif { [nodeType $node2] == "pseudo" } {
	removeLink [getLinkMirror $link]
	removeLink $link
	removeNode [getNodeMirror $node2]
	removeNode $node2
	.panwin.f1.c delete $node2
    } else {
	removeLink $link
    }
    .panwin.f1.c delete $link
    if { $atomic == "atomic" } {
	set changed 1
	updateUndoLog
    }
}

#****f* editor.tcl/removeGUINode
# NAME
#   removeGUINode -- remove node from GUI
# SYNOPSIS
#   renoveGUINode $node_id
# FUNCTION
#   Removes node from GUI. When removing a node from GUI the links
#   connected to that node are also removed.
# INPUTS
#   * node_id -- node id
#****
proc removeGUINode { node } {
    set type [nodeType $node]
    foreach ifc [ifcList $node] {
	set peer [peerByIfc $node $ifc]
	set link [lindex [.panwin.f1.c gettags "link && $node && $peer"] 1]
	removeGUILink $link non-atomic
    }
    if { $type != "pseudo" } {
	removeNode $node
	.panwin.f1.c delete $node
    }
}

#****f* editor.tcl/updateUndoLog
# NAME
#   updateUndoLog -- update the undo log
# SYNOPSIS
#   updateUndoLog
# FUNCTION
#   Updates the undo log. Writes the current configuration to the 
#   undolog array and updates the undolevel variable.
#****
proc updateUndoLog {} {
    upvar 0 ::cf::[set ::curcfg]::undolevel undolevel
    upvar 0 ::cf::[set ::curcfg]::redolevel redolevel
    upvar 0 ::cf::[set ::curcfg]::undolog undolog
    global changed showTree

    if { $changed } {
	global t_undolog
	set t_undolog ""
	dumpCfg string t_undolog
	incr undolevel
	if { $undolevel == 1 } {
	    .menubar.edit entryconfigure "Undo" -state normal
	}
	set undolog($undolevel) $t_undolog
	set redolevel $undolevel
	set changed 0
	if { $showTree } {
	    refreshTopologyTree
	}
    }
}

#****f* editor.tcl/undo
# NAME
#   undo -- undo function
# SYNOPSIS
#   undo 
# FUNCTION
#   Undo the change. Reads the undolog and updates the current
#   configuration. Reduces the value of undolevel.
#****
proc undo {} {
    upvar 0 ::cf::[set ::curcfg]::undolevel undolevel
    upvar 0 ::cf::[set ::curcfg]::undolog undolog
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode

    if {$oper_mode == "edit" && $undolevel > 0} {
	.menubar.edit entryconfigure "Redo" -state normal
	incr undolevel -1
	if { $undolevel == 0 } {
	    .menubar.edit entryconfigure "Undo" -state disabled
	}
	.panwin.f1.c config -cursor watch
	loadCfg $undolog($undolevel)
	switchCanvas none
    }
}

#****f* editor.tcl/redo
# NAME
#   redo -- redo function
# SYNOPSIS
#   redo
# FUNCTION
#   Redo the change if possible (redolevel is greater than 
#   undolevel). Reads the configuration from undolog and
#   updates the current configuration. Increases the value 
#   of undolevel. 
#****
proc redo {} {
    upvar 0 ::cf::[set ::curcfg]::undolevel undolevel
    upvar 0 ::cf::[set ::curcfg]::redolevel redolevel
    upvar 0 ::cf::[set ::curcfg]::undolog undolog
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode

    if {$oper_mode == "edit" && $redolevel > $undolevel} {
	incr undolevel
	if { $undolevel == 1 } {
	    .menubar.edit entryconfigure "Undo" -state normal
	}
	if {$redolevel <= $undolevel} {
	    .menubar.edit entryconfigure "Redo" -state disabled
	}
	.panwin.f1.c config -cursor watch
	loadCfg $undolog($undolevel)
	switchCanvas none
    }
}

#****f* editor.tcl/redrawAll
# NAME
#   redrawAll -- redraw all
# SYNOPSIS
#   redrawAll
# FUNCTION
#   Redraws all the objects on the current canvas.
#****
proc redrawAll {} {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::link_list link_list
    upvar 0 ::cf::[set ::curcfg]::annotation_list annotation_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global background sizex sizey grid
    global showBkgImage showAnnotations showGrid bkgImage
    .bottom.zoom config -text "zoom [expr {int($zoom * 100)}]%"
    set e_sizex [expr {int($sizex * $zoom)}]
    set e_sizey [expr {int($sizey * $zoom)}]
    set border 28
    .panwin.f1.c configure -scrollregion \
	"-$border -$border [expr {$e_sizex + $border}] \
	[expr {$e_sizey + $border}]"

    .panwin.f1.c delete all

    set canvasBkgImage [getCanvasBkg $curcanvas]
    if { $showBkgImage == 1 && "$canvasBkgImage" != ""} {
	set ret [backgroundImage .panwin.f1.c $canvasBkgImage]
	if { "$ret" == 2 } {
	    set background [.panwin.f1.c create rectangle 0 0 $e_sizex $e_sizey \
		    -fill white -tags "background"]
	} else {
	    set background [.panwin.f1.c create rectangle 0 0 $e_sizex $e_sizey \
		    -tags "background"]
	}
    } else {
	set background [.panwin.f1.c create rectangle 0 0 $e_sizex $e_sizey \
	    -fill white -tags "background"]
    }

    if { $showAnnotations == 1 } {
	foreach obj $annotation_list {
	    if { [getNodeCanvas $obj] == $curcanvas } {
		drawAnnotation $obj
	    }
	} 
    }

    # Grid
    set e_grid [expr {int($grid * $zoom)}]
    set e_grid2 [expr {$e_grid * 2}]
    if { $showGrid } {
	for { set x $e_grid } { $x < $e_sizex } { incr x $e_grid } {
	    if { [expr {$x % $e_grid2}] != 0 } {
		if { $zoom > 0.5 } {
		    .panwin.f1.c create line $x 1 $x $e_sizey \
			-fill gray -dash {1 7} -tags "grid"
		}
	    } else {
		.panwin.f1.c create line $x 1 $x $e_sizey -fill gray -dash {1 3} \
		   -tags "grid"
	    }
	}
	for { set y $e_grid } { $y < $e_sizey } { incr y $e_grid } {
	    if { [expr {$y % $e_grid2}] != 0 } {
		if { $zoom > 0.5 } {
		    .panwin.f1.c create line 1 $y $e_sizex $y \
			-fill gray -dash {1 7} -tags "grid"
		}
	    } else {
		.panwin.f1.c create line 1 $y $e_sizex $y -fill gray -dash {1 3} \
		    -tags "grid"
	    }
	}
    }

    .panwin.f1.c lower -withtags background

    foreach node $node_list {
	if { [getNodeCanvas $node] == $curcanvas } {
	    drawNode $node
	}
    }
    foreach link $link_list {
	set nodes [linkPeers $link]
	if { [getNodeCanvas [lindex $nodes 0]] != $curcanvas ||
	    [getNodeCanvas [lindex $nodes 1]] != $curcanvas } {
	    continue
	}
	drawLink $link
	redrawLink $link
	updateLinkLabel $link
    }
    updateIconSize
    .panwin.f1.c config -cursor left_ptr
    raiseAll .panwin.f1.c
}

#****f* editor.tcl/drawNode
# NAME
#   drawNode -- draw a node
# SYNOPSIS
#   drawNode node_id
# FUNCTION
#   Draws the specified node. Draws node's image (router pc
#   host lanswitch frswitch rj45 hub pseudo) and label.
#   The visibility of the label depends on the showNodeLabels
#   variable for all types of nodes and on invisible variable 
#   for pseudo nodes.
# INPUTS
#   * node_id -- node id
#****
proc drawNode { node } {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global showNodeLabels pseudo

    set type [nodeType $node]
    set coords [getNodeCoords $node]
    set x [expr {[lindex $coords 0] * $zoom}]
    set y [expr {[lindex $coords 1] * $zoom}]
    set customIcon [getCustomIcon $node]
    if {[string match "*img*" $customIcon] == 0} {
	global $type
	.panwin.f1.c create image $x $y -image [set $type] -tags "node $node"
    } else {
	global iconSize
	switch $iconSize {
	    normal {
		set icon_data [getImageData $customIcon]
		image create photo img_$customIcon -data $icon_data
		.panwin.f1.c create image $x $y -image img_$customIcon -tags "node $node"
	    }
	    small {
		set icon_data [getImageData $customIcon]
		image create photo img_$customIcon -data $icon_data
		set img_$customIcon [image% img_$customIcon 70 $customIcon]
		.panwin.f1.c create image $x $y -image [set img_$customIcon] -tags "node $node"
	    }
	}
    }
    set coords [getNodeLabelCoords $node]
    set x [expr {[lindex $coords 0] * $zoom}]
    set y [expr {[lindex $coords 1] * $zoom}]
    if { [nodeType $node] != "pseudo" } {
	set labelstr1 [getNodeName $node];
	set labelstr2 [getNodePartition $node];
	set l [format "%s\n%s" $labelstr1 $labelstr2];
	set label [.panwin.f1.c create text $x $y -fill blue \
	    -text "$l" \
	    -tags "nodelabel $node"]
    } else {
	set pnode [getNodeName $node]
	set pcanvas [getNodeCanvas $pnode]
	set ifc [ifcByPeer $pnode [getNodeMirror $node]]
	if { $pcanvas != $curcanvas } {
	    set label [.panwin.f1.c create text $x $y -fill blue \
		-text "[getNodeName $pnode]:$ifc
@[getCanvasName $pcanvas]" \
		-tags "nodelabel $node" -justify center]
	} else {
	    set label [.panwin.f1.c create text $x $y -fill blue \
		-text "[getNodeName $pnode]:$ifc" \
		-tags "nodelabel $node" -justify center]
	}
    }
    if { $showNodeLabels == 0} {
	.panwin.f1.c itemconfigure $label -state hidden
    }
    # XXX Invisible pseudo-node labels
    global invisible
    if { $invisible == 1 && [nodeType $node] == "pseudo" } {
	.panwin.f1.c itemconfigure $label -state hidden
    }
    if {[nodeType $node] == "cloud"} {
	setCloudParts $node 1
    }
}

#****f* editor.tcl/drawLink
# NAME
#   drawLink -- draw a link
# SYNOPSIS
#   drawLink link_id
# FUNCTION
#   Draws the specified link. An arrow is displayed for links 
#   connected to pseudo nodes. If the variable invisible
#   is specified link connecting a pseudo node stays hidden. 
# INPUTS
#   * link_id -- link id
#****
proc drawLink { link } {
    set nodes [linkPeers $link]
    set lnode1 [lindex $nodes 0]
    set lnode2 [lindex $nodes 1]
    set lwidth [getLinkWidth $link]
    if { [getLinkMirror $link] != "" } {
	set newlink [.panwin.f1.c create line 0 0 0 0 \
	    -fill [getLinkColor $link] -width $lwidth \
	    -tags "link $link $lnode1 $lnode2" -arrow both]
    } else {
	set newlink [.panwin.f1.c create line 0 0 0 0 \
	    -fill [getLinkColor $link] -width $lwidth \
	    -tags "link $link $lnode1 $lnode2"]
    }
    # XXX Invisible pseudo-liks
    global invisible
    if { $invisible == 1 && [getLinkMirror $link] != "" } {
	.panwin.f1.c itemconfigure $link -state hidden
    }
    .panwin.f1.c raise $newlink background
    set newlink [.panwin.f1.c create line 0 0 0 0 \
	-fill white -width [expr {$lwidth + 4}] \
	-tags "link $link $lnode1 $lnode2"]
    .panwin.f1.c raise $newlink background

    set ang [calcAngle $link]

    .panwin.f1.c create text 0 0 -tags "linklabel $link" -justify center -angle $ang -anchor center
    .panwin.f1.c create text 0 0 -tags "interface $lnode1 $link" -justify center -angle $ang
    .panwin.f1.c create text 0 0 -tags "interface $lnode2 $link" -justify center -angle $ang

    .panwin.f1.c raise linklabel "link || background"
    .panwin.f1.c raise interface "link || linklabel || background"
}

#****f* editor.tcl/chooseIfName
# NAME
#   chooseIfName -- choose interface name
# SYNOPSIS
#   set ifcName [chooseIfName $lnode1 $lnode2]
# FUNCTION
#   Choose intreface name. The name can be:
#   * eth -- for interface connecting pc, host and router  
#   * e -- for interface connecting hub and lanswitch
#   * f -- for interface connecting frswitch
# INPUTS
#   * link_id -- link id
# RESULT
#   * ifcName -- the name of the interface
#****
proc chooseIfName { lnode1 lnode2 } {

    if { [nodeType $lnode1] == "frswitch" } {
		return f
    } else {
		return [[nodeType $lnode1].ifcName]
    }
}

#****f* editor.tcl/listLANNodes
# NAME
#   listLANNodes -- list LAN nodes
# SYNOPSIS
#   set l2peers [listLANNodes $l2node $l2peers]
# FUNCTION
#   Recursive function for finding all link layer nodes that are 
#   connected to node l2node. Returns the list of all link layer 
#   nodes that are on the same LAN as l2node.
# INPUTS
#   * l2node -- node id of a link layer node
#   * l2peers -- old link layer nodes on the same LAN
# RESULT
#   * l2peers -- new link layer nodes on the same LAN
#****
proc listLANnodes { l2node l2peers } {
    lappend l2peers $l2node
    foreach ifc [ifcList $l2node] {
	set peer [logicalPeerByIfc $l2node $ifc]
	if {[[typemodel $peer].layer] == "LINK" &&  [nodeType $peer] != "rj45"} {
	    if { [lsearch $l2peers $peer] == -1 } {
		set l2peers [listLANnodes $peer $l2peers]
	    }
	}
    }
    return $l2peers
}

#****f* editor.tcl/calcAngle
# NAME
#   calcAngle -- calculate link angle
# SYNOPSIS
#   calcAngle $link
# FUNCTION
#   Calculates the angle of the link that connects 2 nodes.
#   Used to calculate the rotation angle of link and interface labels.
# INPUTS
#   * link -- link which rotation angle needs to be calculated.
#****
proc calcAngle { link } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    set nodes [linkPeers $link]
    set lnode1 [lindex $nodes 0]
    set lnode2 [lindex $nodes 1]
    set coords [getNodeCoords $lnode1]
    set x1 [expr {[lindex $coords 0]*$zoom}]
    set y1 [expr {[lindex $coords 1]*$zoom}]
    set coords [getNodeCoords $lnode2]
    set x2 [expr {[lindex $coords 0]*$zoom}]
    set y2 [expr {[lindex $coords 1]*$zoom}]
    if { [expr $x2 - $x1] == 0 } {
	set arad 0
    } else {
	set arad [expr {atan(($y2-$y1)/($x2-$x1))}]
    }
    set ang [expr {$arad*180/3.14159}]
    if {$ang < 0} {
	set ang [expr {$ang+360}]
    }
    set ang [expr {360-$ang}]
    if {$ang > 255 && $ang < 320 || $ang > 40 && $ang < 105} {
	set ang 0
    }
    return $ang
}

#****f* editor.tcl/calcDxDy
# NAME
#   calcDxDy -- calculate dx and dy
# SYNOPSIS
#   calcDxDy $node
# FUNCTION
#   Calculates dx and dy variables of the calling function.
# INPUTS
#   * node -- node id of a node whose dx and dy coordinates are 
#   calculated
#****
proc calcDxDy { node } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    upvar dx x
    upvar dy y

    if { $zoom > 1.0 } {
	set x 1
	set y 1
	return
    }

    if { [nodeType $node] == "frswitch" } {
	set x [expr {1.8 / $zoom}]
	set y [expr {1.8 / $zoom}]
#    } elseif {[nodeType $node] == "pseudo"} {
    } else {
	set x [lindex [[nodeType $node].calcDxDy] 0]
	set y [lindex [[nodeType $node].calcDxDy] 1]
    }
}

#****f* editor.tcl/updateIfcLabel
# NAME
#   updateIfcLabel -- update interface label
# SYNOPSIS
#   updateIfcLabel $lnode1 $lnode2
# FUNCTION
#   Updates the interface label, including interface name,
#   interface state (* for interfaces that are down), IPv4
#   address and IPv6 address.
# INPUTS
#   * lnode1 -- node id of a node where the interface resides
#   * lnode2 -- node id of the node that is connected by this 
#   interface. 
#****
proc updateIfcLabel { lnode1 lnode2 } {
    global showIfNames showIfIPaddrs showIfIPv6addrs

    set link [lindex [.panwin.f1.c gettags "link && $lnode1 && $lnode2"] 1]
    set ifc [ifcByPeer $lnode1 $lnode2]
    set ifipv4addr [getIfcIPv4addr $lnode1 $ifc]
    set ifipv6addr [getIfcIPv6addr $lnode1 $ifc]
    if { $ifc == 0 } {
	set ifc ""
    }
    set labelstr ""
    if { $showIfNames } {
	lappend labelstr "$ifc"
    }
    if { $showIfIPaddrs && $ifipv4addr != "" } {
	lappend labelstr "$ifipv4addr"
    }
    if { $showIfIPv6addrs && $ifipv6addr != "" } {
	lappend labelstr "$ifipv6addr"
    }
    if { [getIfcOperState $lnode1 $ifc] == "down" } {
	set str "*"
    } else {
	set str ""
    }
    foreach elem $labelstr {
	if {$str == "" || $str == "*"} {
	    set str "$str[set elem]"
	} else {
	    set str "$str\r[set elem]"
	}
    }
    set ang [calcAngle $link]
    .panwin.f1.c itemconfigure "interface && $lnode1 && $link" \
	-text $str -angle $ang
}

#****f* editor.tcl/updateLinkLabel
# NAME
#   updateLinkLabel -- update link label
# SYNOPSIS
#   updateLinkLabel $link
# FUNCTION
#   Updates the link label, including link bandwidth, link delay,
#   BER and duplicate values.
# INPUTS
#   * link -- link id of the link whose labels are updated.
#****
proc updateLinkLabel { link } {
    global showLinkLabels linkJitterConfiguration 

    set labelstr ""
    set bwstr "[getLinkBandwidthString $link]"
    set delstr [getLinkDelayString $link]
    set ber [getLinkBER $link]
    set dup [getLinkDup $link]
    set jitter [concat [getLinkJitterUpstream $link] [getLinkJitterDownstream $link]]
    if { "$bwstr" != "" } {
	lappend labelstr $bwstr
    }
    if { "$delstr" != "" } {
	lappend labelstr $delstr
    }
    if { $jitter != "" && $linkJitterConfiguration == 1 } {
	lappend labelstr "jitter"
    }
    if { "$ber" != "" } {
	lappend labelstr "ber=$ber"
    }
    if { "$dup" != "" } {
	lappend labelstr "dup=$dup%"
    }
    set str ""
    foreach elem $labelstr {
	if {$str == ""} {
	    set str "$str[set elem]"
	} else {
	    set str "$str\r[set elem]"
	}
    }
    set ang [calcAngle $link]
    .panwin.f1.c itemconfigure "linklabel && $link" -text $str -angle $ang
    if { $showLinkLabels == 0} {
	.panwin.f1.c itemconfigure "linklabel && $link" -state hidden
    }
}

#****f* editor.tcl/redrawAllLinks
# NAME
#   redrawAllLinks -- redraw all links
# SYNOPSIS
#   redrawAllLinks
# FUNCTION
#   Redraws all links on the current canvas.
#****
proc redrawAllLinks {} {
    upvar 0 ::cf::[set ::curcfg]::link_list link_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas

    foreach link $link_list {
	set nodes [linkPeers $link]
	if { [getNodeCanvas [lindex $nodes 0]] != $curcanvas ||
	    [getNodeCanvas [lindex $nodes 1]] != $curcanvas } {
	    continue
	}
	redrawLink $link
    }
}

#****f* editor.tcl/redrawLink
# NAME
#   redrawLink -- redraw a link
# SYNOPSIS
#   redrawLink $link
# FUNCTION
#   Redraws the specified link.
# INPUTS
#   * link -- link id
#****
proc redrawLink { link } {
    set limages [.panwin.f1.c find withtag "link && $link"]
    set limage1 [lindex $limages 0]
    set limage2 [lindex $limages 1]
    set tags [.panwin.f1.c gettags $limage1]
    set link [lindex $tags 1]
    set lnode1 [lindex $tags 2]
    set lnode2 [lindex $tags 3]

    set coords1 [.panwin.f1.c coords "node && $lnode1"]
    set coords2 [.panwin.f1.c coords "node && $lnode2"]
    set x1 [lindex $coords1 0]
    set y1 [lindex $coords1 1]
    set x2 [lindex $coords2 0]
    set y2 [lindex $coords2 1]

    .panwin.f1.c coords $limage1 $x1 $y1 $x2 $y2
    .panwin.f1.c coords $limage2 $x1 $y1 $x2 $y2

    set lx [expr {0.5 * ($x1 + $x2)}]
    set ly [expr {0.5 * ($y1 + $y2)}]
    .panwin.f1.c coords "linklabel && $link" $lx $ly

    set nx [expr {sqrt (($x1 - $x2) * ($x1 - $x2) + \
	($y1 - $y2) * ($y1 - $y2)) * 0.0125}]
    if { $nx < 1 } {
	set nx 1
    }

    set ny [expr {sqrt (($x1 - $x2) * ($x1 - $x2) + \
	($y1 - $y2) * ($y1 - $y2)) * 0.010}]
    if { $ny < 1 } {
	set ny 1
    }
    
    calcDxDy $lnode1
    set lx [expr {($x1 * ($nx * $dx - 1) + $x2) / $nx / $dx}]
    set ly [expr {($y1 * ($ny * $dy - 1) + $y2) / $ny / $dy}]
    .panwin.f1.c coords "interface && $lnode1 && $link" $lx $ly
    updateIfcLabel $lnode1 $lnode2

    calcDxDy $lnode2
    set lx [expr {($x1 + $x2 * ($nx * $dx - 1)) / $nx / $dx}]
    set ly [expr {($y1 + $y2 * ($ny * $dy - 1)) / $ny / $dy}]
    .panwin.f1.c coords "interface && $lnode2 && $link" $lx $ly
    updateIfcLabel $lnode2 $lnode1
}

#****f* editor.tcl/splitGUILink
# NAME
#   splitGUILink -- splits a link
# SYNOPSIS
#   splitGUILink $link
# FUNCTION
#   Splits the link and draws new links and new pseudo nodes 
#   on the canvas.
# INPUTS
#   * link -- link id
#****
proc splitGUILink { link } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global changed

    set peer_nodes [linkPeers $link]
    set new_nodes [splitLink $link pseudo]
    set orig_node1 [lindex $peer_nodes 0]
    set orig_node2 [lindex $peer_nodes 1]
    set new_node1 [lindex $new_nodes 0]
    set new_node2 [lindex $new_nodes 1]
    set new_link1 [linkByPeers $orig_node1 $new_node1]
    set new_link2 [linkByPeers $orig_node2 $new_node2]
    setLinkMirror $new_link1 $new_link2
    setLinkMirror $new_link2 $new_link1
    setNodeMirror $new_node1 $new_node2
    setNodeMirror $new_node2 $new_node1
    setNodeName $new_node1 $orig_node2
    setNodeName $new_node2 $orig_node1

    set x1 [lindex [getNodeCoords $orig_node1] 0]
    set y1 [lindex [getNodeCoords $orig_node1] 1]
    set x2 [lindex [getNodeCoords $orig_node2] 0]
    set y2 [lindex [getNodeCoords $orig_node2] 1]

    setNodeCoords $new_node1 \
	"[expr {($x1 + 0.4 * ($x2 - $x1)) / $zoom}] \
	[expr {($y1 + 0.4 * ($y2 - $y1)) / $zoom}]"
    setNodeCoords $new_node2 \
	"[expr {($x1 + 0.6 * ($x2 - $x1)) / $zoom}] \
	[expr {($y1 + 0.6 * ($y2 - $y1)) / $zoom}]"
    setNodeLabelCoords $new_node1 [getNodeCoords $new_node1]
    setNodeLabelCoords $new_node2 [getNodeCoords $new_node2]

    set changed 1
    updateUndoLog
    redrawAll
}

#****f* editor.tcl/selectNode
# NAME
#   selectNode -- select node 
# SYNOPSIS
#   selectNode $c $obj
# FUNCTION
#   Crates the selecting box around the specified canvas
#   object.
# INPUTS
#   * c -- tk canvas
#   * obj -- tk canvas object tag id
#****
proc selectNode { c obj } {
    set node [lindex [$c gettags $obj] 1]

    if { $node == "" } {
	return
    }
    $c addtag selected withtag "node && $node"
    if { [nodeType $node] == "pseudo" } {
	set bbox [$c bbox "nodelabel && $node"]
    } elseif { [nodeType $node] == "rectangle" } {
	$c addtag selected withtag "rectangle && $node"
	set bbox [$c bbox "rectangle && $node"]
    } elseif { [nodeType $node] == "text" } {
	$c addtag selected withtag "text && $node"
	set bbox [$c bbox "text && $node"]
    } elseif { [nodeType $node] == "oval" } {
	$c addtag selected withtag "oval && $node"
	set bbox [$c bbox "oval && $node"]
    } elseif { [nodeType $node] == "freeform" } {
	$c addtag selected withtag "freeform && $node"
	set bbox [$c bbox "freeform && $node"]
    } else {
	set bbox [$c bbox "node && $node"]
    }
    set bx1 [expr {[lindex $bbox 0] - 2}]
    set by1 [expr {[lindex $bbox 1] - 2}]
    set bx2 [expr {[lindex $bbox 2] + 1}]
    set by2 [expr {[lindex $bbox 3] + 1}]
    $c delete -withtags "selectmark && $node"
    $c create line $bx1 $by1 $bx2 $by1 $bx2 $by2 $bx1 $by2 $bx1 $by1 \
	-dash {6 4} -fill black -width 1 -tags "selectmark $node"
}

#****f* editor.tcl/selectAllObjects
# NAME
#   selectAllObjects -- select all objects on the canvas
# SYNOPSIS
#   selectAllObjects
# FUNCTION
#   Select all object on the canvas.
#****
proc selectAllObjects { } {
    foreach obj [.panwin.f1.c find withtag "node || text || oval || rectangle \
	|| freeform"] {
	selectNode .panwin.f1.c $obj
    }
}

#****f* editor.tcl/selectNodes
# NAME
#   selectNodes -- select nodes
# SYNOPSIS
#   selectNodes $nodelist
# FUNCTION
#   Select all nodes in a list.
# INPUTS
#   * nodelist -- list of nodes to select.
#****
proc selectNodes { nodelist } {
    foreach node $nodelist {
	selectNode .panwin.f1.c [.panwin.f1.c find withtag \
	    "(node || text || oval || rectangle || freeform) && $node"]
    }
}

#****f* editor.tcl/selectedNodes
# NAME
#   selectedNodes -- get selected nodes
# SYNOPSIS
#   selectedNodes
# FUNCTION
#   Gets selected nodes and returns them as a list.
# RESULT
#   * selected -- object list of selected nodes.
#****
proc selectedNodes {} {
    set selected {}
    foreach obj [.panwin.f1.c find withtag "node && selected"] {
	lappend selected [lindex [.panwin.f1.c gettags $obj] 1]
    }
    foreach obj [.panwin.f1.c find withtag "oval && selected"] {
	lappend selected [lindex [.panwin.f1.c gettags $obj] 1]
    }
    foreach obj [.panwin.f1.c find withtag "rectangle && selected"] {
	lappend selected [lindex [.panwin.f1.c gettags $obj] 1]
    }
    foreach obj [.panwin.f1.c find withtag "text && selected"] {
	lappend selected [lindex [.panwin.f1.c gettags $obj] 1]
    }
    foreach obj [.panwin.f1.c find withtag "freeform && selected"] {
	lappend selected [lindex [.panwin.f1.c gettags $obj] 1]
    }
    return $selected
}

#****f* editor.tcl/selectedRealNodes
# NAME
#   selectedRealNodes -- get selected real nodes
# SYNOPSIS
#   selectedRealNodes
# FUNCTION
#   Gets selected real nodes and returns them as a list.
# RESULT
#   * selected -- object list of selected real nodes.
#****
proc selectedRealNodes {} {
    set selected {}
    foreach obj [.panwin.f1.c find withtag "node && selected"] {
	set node [lindex [.panwin.f1.c gettags $obj] 1]
	if { [getNodeMirror $node] != "" ||
	    [nodeType $node] == "rj45" } {
	    continue
	}
	lappend selected $node
    }
    return $selected
}

#****f* editor.tcl/selectAdjacent
# NAME
#   selectAdjacent -- select adjacent nodes
# SYNOPSIS
#   selectAdjacent
# FUNCTION
#   Finds all adjacent nodes and selects them.
#****
proc selectAdjacent {} {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas

    set selected [selectedNodes]
    set adjacent {}
    foreach node $selected {
	foreach ifc [ifcList $node] {
	    set peer [peerByIfc $node $ifc]
	    if { [getNodeMirror $peer] != "" } {
		return
	    }
	    if { [lsearch $adjacent $peer] < 0 } {
		lappend adjacent $peer
	    }
	}
    }
    selectNodes $adjacent
}

#****f* editor.tcl/button3link
# NAME
#   button3link 
# SYNOPSIS
#   button3link $c $x $y
# FUNCTION
#   This procedure is called when a right mouse button is 
#   clicked on the canvas. If there is a link on the place of
#   mouse click this procedure creates and configures a popup
#   menu. The options in the menu are:
#   * Configure -- configure the link
#   * Delete -- delete the link
#   * Split -- split the link
#   * Merge -- this option is active only if the link is previously
#   been split, by this action the link is merged.
# INPUTS
#   * c -- tk canvas
#   * x -- x coordinate for popup menu
#   * y -- y coordinate for popup menu
#****
proc button3link { c x y } {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::canvas_list canvas_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode

    set link [lindex [$c gettags {link && current}] 1]
    if { $link == "" } {
	set link [lindex [$c gettags {linklabel && current}] 1]
	if { $link == "" } {
	    return
	}
    }

    .button3menu delete 0 end

    #
    # Configure link
    #
    .button3menu add command -label "Configure" \
	-command "linkConfigGUI $c $link"

    #
    # Clear link configuration 
    #
    .button3menu add command -label "Clear all settings" \
	-command "linkResetConfig $link"

    global linkJitterConfiguration
    if  { $linkJitterConfiguration } {
	#
	# Edit link jitter
	#
	.button3menu add command -label "Edit link jitter" \
	    -command "linkJitterConfigGUI $c $link"
	#
	# Reset link jitter
	#
	.button3menu add command -label "Clear link jitter" \
	    -command "linkJitterReset $link"
    }

    #
    # Delete link
    #
    if { $oper_mode != "exec" } {
	.button3menu add command -label "Delete" \
	    -command "removeGUILink $link atomic"
    } else {
	.button3menu add command -label "Delete" \
	    -state disabled
    }

    #
    # Split link
    #
    if { $oper_mode != "exec" && [getLinkMirror $link] == "" } {
	.button3menu add command -label "Split" \
	    -command "splitGUILink $link"
    } else {
	.button3menu add command -label "Split" \
	    -state disabled
    }

    #
    # Merge two pseudo nodes / links
    #
    if { $oper_mode != "exec" && [getLinkMirror $link] != "" &&
	[getNodeCanvas [getNodeMirror [lindex [linkPeers $link] 1]]] ==
	$curcanvas } {
	.button3menu add command -label "Merge" \
	    -command "mergeGUINode [lindex [linkPeers $link] 1]"
    } else {
	.button3menu add command -label "Merge" -state disabled
    }

    set x [winfo pointerx .]
    set y [winfo pointery .]
    tk_popup .button3menu $x $y
}

#****f* editor.tcl/movetoCanvas
# NAME
#   movetoCanvas -- move to canvas 
# SYNOPSIS
#   movetoCanvas $canvas
# FUNCTION
#   This procedure moves all the nodes selected in the GUI to
#   the specified canvas.
# INPUTS
#   * canvas -- canvas id.
#****
proc movetoCanvas { canvas } {
    global changed

    set selected_nodes [selectedNodes]
    foreach node $selected_nodes {
	setNodeCanvas $node $canvas
	set changed 1
    }
    foreach obj [.panwin.f1.c find withtag "linklabel"] {
	set link [lindex [.panwin.f1.c gettags $obj] 1]
	set link_peers [linkPeers $link]
	set peer1 [lindex $link_peers 0]
	set peer2 [lindex $link_peers 1]
	set peer1_in_selected [lsearch $selected_nodes $peer1]
	set peer2_in_selected [lsearch $selected_nodes $peer2]
	if { ($peer1_in_selected == -1 && $peer2_in_selected != -1) ||
	    ($peer1_in_selected != -1 && $peer2_in_selected == -1) } {
	    if { [nodeType $peer2] == "pseudo" } {
		setNodeCanvas $peer2 $canvas
		if { [getNodeCanvas [getNodeMirror $peer2]] == $canvas } {
		    mergeLink $link
		}
		continue
	    }
	    set new_nodes [splitLink $link pseudo]
	    set new_node1 [lindex $new_nodes 0]
	    set new_node2 [lindex $new_nodes 1]
	    setNodeMirror $new_node1 $new_node2
	    setNodeMirror $new_node2 $new_node1
	    setNodeName $new_node1 $peer2
	    setNodeName $new_node2 $peer1
	    set link1 [linkByPeers $peer1 $new_node1]
	    set link2 [linkByPeers $peer2 $new_node2]
	    setLinkMirror $link1 $link2
	    setLinkMirror $link2 $link1
	}
    }
    updateUndoLog
    redrawAll
}

#****f* editor.tcl/mergeGUINode
# NAME
#   mergeGUINode -- merge GUI node
# SYNOPSIS
#   mergeGUINode $node
# FUNCTION
#   This procedure removes the specified pseudo node as well
#   as it's mirror copy. Also this procedure removes the
#   pseudo links and reestablish the original link between
#   the non-pseudo nodes.
# INPUTS
#   * node -- node id of a pseudo node.
#****
proc mergeGUINode { node } {
    set link [lindex [linkByIfc $node [ifcList $node]] 0]
    mergeLink $link
    redrawAll
}

#****f* editor.tcl/button3node
# NAME
#   button3node
# SYNOPSIS
#   button3node $c $x $y
# FUNCTION
#   This procedure is called when a right mouse button is 
#   clicked on the canvas. If there is a node on the place of
#   mouse click this procedure creates and configures a popup
#   menu. The options in the menu are:
#   * Configure -- configure the node
#   * Create link to -- create a link to any available node,
#   it can be on the same canvas or on a different canvas.
#   * Move to -- move to some other canvas
#   * Merge -- this option is available only for pseudo nodes
#   that have mirror nodes on the same canvas (Pseudo nodes
#   created by splitting a link).
#   * Delete -- delete the node
#   * Shell window -- specifies the shell window to open in 
#   exec mode. This option is available only to nodes on a 
#   network layer
#   * Wireshark -- opens a Wireshark program for the specified 
#   node and the specified interface. This option is available 
#   only for network layer nodes in exec mode.
# INPUTS
#   * c -- tk canvas
#   * x -- x coordinate for popup menu
#   * y -- y coordinate for popup menu
#****
proc button3node { c x y } {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::canvas_list canvas_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode
    upvar 0 ::cf::[set ::curcfg]::eid eid

    set node [lindex [$c gettags {node && current}] 1]
    if { $node == "" } {
	set node [lindex [$c gettags {nodelabel && current}] 1]
	if { $node == "" } {
	    return
	}
    }
    set mirror_node [getNodeMirror $node]

    if { [$c gettags "node && $node && selected"] == "" } {
	$c dtag node selected
	$c delete -withtags selectmark
	selectNode $c [$c find withtag "current"]
    }

    .button3menu delete 0 end

    #
    # Select adjacent
    #
    if { [nodeType $node] != "pseudo" } {
	.button3menu add command -label "Select adjacent" \
	    -command "selectAdjacent"
    } else {
	.button3menu add command -label "Select adjacent" \
	    -command "selectAdjacent" -state disabled
    }

    #
    # Configure node
    #
    if { [nodeType $node] != "pseudo" } {
	.button3menu add command -label "Configure" \
	    -command "nodeConfigGUI $c $node"
    } else {
	.button3menu add command -label "Configure" \
	    -command "nodeConfigGUI $c $node" -state disabled
    }
    
    #
    # Node icon preferences
    #   
    .button3menu.sett delete 0 end
    if { [nodeType $node] != "pseudo" } {
	.button3menu add cascade -label "Settings" \
	    -menu .button3menu.sett
    } else {
	.button3menu add cascade -label "Settings" \
	    -menu .button3menu.sett -state disabled
    }
    .button3menu.sett add command -label "Remove IPv4 addresses" \
	    -command "removeIPv4nodes"
    .button3menu.sett add command -label "Remove IPv6 addresses" \
	    -command "removeIPv6nodes"
    if { $oper_mode == "exec" } {
	.button3menu.sett add command -label "Fetch Node Configurations" \
	    -command "fetchNodeConfiguration"
    } else {
        .button3menu.sett add command -label "Fetch Node Configurations" \
	    -state disabled
    } 

    #
    # Node icon preferences
    #   
    .button3menu.icon delete 0 end
    if { [nodeType $node] != "pseudo" } {
	.button3menu add cascade -label "Node icon" \
	    -menu .button3menu.icon
    } else {
	.button3menu add cascade -label "Node icon" \
	    -menu .button3menu.icon -state disabled
    }
    .button3menu.icon add command -label "Change node icons" \
	    -command "changeIconPopup"
    .button3menu.icon add command -label "Set default icons" \
	    -command "setDefaultIcon"

    #
    # Create a new link - can be between different canvases
    #
    .button3menu.connect delete 0 end
    if { $oper_mode == "exec" || [nodeType $node] == "pseudo" } {
	.button3menu add cascade -label "Create link to" \
	    -menu .button3menu.connect -state disabled
    } else {
	.button3menu add cascade -label "Create link to" \
	    -menu .button3menu.connect
    }
    destroy .button3menu.connect.selected
    menu .button3menu.connect.selected -tearoff 0
    .button3menu.connect add cascade -label "Selected" \
	-menu .button3menu.connect.selected
    .button3menu.connect.selected add command \
	-label "Chain" -command "P \[selectedRealNodes\]"
    .button3menu.connect.selected add command \
	-label "Star" \
	-command "Kb \[lindex \[selectedRealNodes\] 0\] \
	\[lrange \[selectedNodes\] 1 end\]"
    .button3menu.connect.selected add command \
	-label "Cycle" -command "C \[selectedRealNodes\]"
    .button3menu.connect.selected add command \
	-label "Clique" -command "K \[selectedRealNodes\]"
    .button3menu.connect.selected add command \
	-label "Random" -command "R \[selectedRealNodes\] \
	\[expr \[llength \[selectedRealNodes\]\] - 1\]"
    .button3menu.connect add separator
    foreach canvas $canvas_list {
	destroy .button3menu.connect.$canvas
	menu .button3menu.connect.$canvas -tearoff 0
	.button3menu.connect add cascade -label [getCanvasName $canvas] \
	    -menu .button3menu.connect.$canvas
    }
    foreach peer_node $node_list {
	set canvas [getNodeCanvas $peer_node]
	if { $node != $peer_node && [nodeType $node] != "rj45" &&
	    [lsearch {pseudo rj45} [nodeType $peer_node]] < 0 &&
	    [ifcByLogicalPeer $node $peer_node] == "" } {
	    .button3menu.connect.$canvas add command \
		-label [getNodeName $peer_node] \
		-command "newGUILink $node $peer_node"
	} elseif { [nodeType $peer_node] != "pseudo" } {
	    .button3menu.connect.$canvas add command \
		-label [getNodeName $peer_node] \
		-state disabled
	}
    }

    #
    # IPv4 autorenumber
    #
    if { $oper_mode == "exec" || [[typemodel $node].layer] == "LINK" } {
	.button3menu add command -label "IPv4 autorenumber" \
	    -state disabled
    } else {
	.button3menu add command -label "IPv4 autorenumber" \
	    -command { 
		global IPv4autoAssign
		set IPv4autoAssign 1
		changeAddressRange 
		set IPv4autoAssign 0
	    }
    }

    #
    # IPv6 autorenumber
    #
    if { $oper_mode == "exec" || [[typemodel $node].layer] == "LINK" } {
	.button3menu add command -label "IPv6 autorenumber" \
	    -state disabled
    } else {
	.button3menu add command -label "IPv6 autorenumber" \
	    -command {
		global IPv6autoAssign
		set IPv6autoAssign 1
		changeAddressRange6 
		set IPv6autoAssign 0
	    }
    }

    #
    # Move to another canvas
    #
    .button3menu.moveto delete 0 end
    if { $oper_mode == "exec" || [nodeType $node] == "pseudo" } {
	.button3menu add cascade -label "Move to" \
	    -menu .button3menu.moveto -state disabled
    } else {
	.button3menu add cascade -label "Move to" \
	    -menu .button3menu.moveto
    }
    .button3menu.moveto add command -label "Canvas:" -state disabled
    foreach canvas $canvas_list {
	if { $canvas != $curcanvas } {
	    .button3menu.moveto add command \
		-label [getCanvasName $canvas] \
		-command "movetoCanvas $canvas"
	} else {
	    .button3menu.moveto add command \
		-label [getCanvasName $canvas] -state disabled
	}
    }

    #
    # Merge two pseudo nodes / links
    #
    if { $oper_mode != "exec" && [nodeType $node] == "pseudo" && \
	[getNodeCanvas $mirror_node] == $curcanvas } {
	.button3menu add command -label "Merge" \
	    -command "mergeGUINode $node"
    } else {
	.button3menu add command -label "Merge" -state disabled
    }

    #
    # Delete selection
    #
    if { $oper_mode != "exec" } {
	.button3menu add command -label "Delete" -command deleteSelection
    } else {
	.button3menu add command -label "Delete" -state disabled
    }

    #
    # Shell selection
    #
    .button3menu.shell delete 0 end
    if {$oper_mode == "exec" && [[typemodel $node].virtlayer] == "VIMAGE"} {
	.button3menu add cascade -label "Shell window" \
	    -menu .button3menu.shell
	foreach cmd [existingShells [[typemodel $node].shellcmds] $node] {
	    .button3menu.shell add command -label "[lindex [split $cmd /] end]" \
		-command "spawnShell $node $cmd"
	}
    } else {
	.button3menu add cascade -label "Shell window" \
	    -menu .button3menu.shell -state disabled
    }

    .button3menu.ethereal delete 0 end
    if {$oper_mode == "exec" && [[typemodel $node].virtlayer] == "VIMAGE"} {
	global debug
	#
	# Wireshark
	#
	.button3menu add cascade -label "Wireshark" \
	    -menu .button3menu.ethereal
	foreach ifc [ifcList $node] {
	    set label "$ifc"
	    if { [getIfcIPv4addr $node $ifc] != "" } {
		set label "$label ([getIfcIPv4addr $node $ifc])"
	    }
	    if { [getIfcIPv6addr $node $ifc] != "" } {
		set label "$label ([getIfcIPv6addr $node $ifc])"
	    }
	    set wireshark [ catch { exec jexec $eid.$node which wireshark } ] 
	    if {[file exists /usr/local/bin/startxcmd] == 1 && $wireshark == 0} {
		.button3menu.ethereal add command -label $label \
		    -command "nexec startxcmd [getNodeName $node]@$eid wireshark -ki $ifc &"
	    } else {
		.button3menu.ethereal add command -label $label \
		    -command "startethereal $node $ifc"
	    }
	}
	set firefox [ catch { exec jexec $eid.$node which firefox } ] 
	if {[file exists /usr/local/bin/startxcmd] == 1 && $firefox == 0 } {
	    .button3menu add command -label "Web Browser" \
		-command "nexec startxcmd [getNodeName $node]@$eid firefox -no-remote about:blank &"
	} else {
	    .button3menu add command -label "Web Browser" \
		-state disabled
	}
    } else {
	.button3menu add cascade -label "Wireshark" \
	    -menu .button3menu.ethereal -state disabled
	.button3menu add command -label "Web Browser" \
	    -state disabled
    }
    if {$oper_mode == "exec" && [info procs [typemodel $node].start] != "" \
	&& [info procs [typemodel $node].shutdown] != ""} {
	.button3menu add command -label Start \
	    -command "[typemodel $node].start $eid $node"
	.button3menu add command -label Stop \
	    -command "[typemodel $node].shutdown $eid $node" 
    } else {
	.button3menu add command -label Start \
	    -command "[typemodel $node].start $eid $node" -state disabled
	.button3menu add command -label Stop \
	    -command "[typemodel $node].shutdown $eid $node" -state disabled 
    }

    #
    # Finally post the popup menu on current pointer position
    #
    set x [winfo pointerx .]
    set y [winfo pointery .]
    tk_popup .button3menu $x $y
}

#****f* editor.tcl/existingShells
# NAME
#   existingShells -- check which shells exist in a node
# SYNOPSIS
#   existingShells $shells $node
# FUNCTION
#   This procedure checks which of the provided shells are available
#   in a running node. 
# INPUTS
#   * shells -- list of shells.
#   * node -- node id of the node for which the check is performed.
#****
proc existingShells { shells node } {
    upvar 0 ::cf::[set ::curcfg]::eid eid

    set cmd "jexec $eid.$node which $shells"

    catch {eval exec $cmd} res

    return $res
}

#****f* editor.tcl/spawnShell
# NAME
#   spawnShell -- spawn shell
# SYNOPSIS
#   spawnShell $node $cmd
# FUNCTION
#   This procedure spawns a new shell for a specified node.
#   The shell is specified in cmd parameter.
# INPUTS
#   * node -- node id of the node for which the shell is spawned.
#   * cmd -- the path to the shell.
#****
proc spawnShell { node cmd } {
    upvar 0 ::cf::[set ::curcfg]::eid eid
    global gui_unix

    set node_id $eid\.$node

    nexec xterm -sb -rightbar \
	-T "IMUNES: [getNodeName $node] (console) [lindex [split $cmd /] end]" \
	-e "jexec $node_id script -q /dev/null $cmd" &
}

#****f* editor.tcl/spawnShellExec
# NAME
#   spawnShellExec -- spawn shell in exec mode on double click 
# SYNOPSIS
#   spawnShellExec $cmd
# FUNCTION
#   This procedure spawns a new shell on a selected and current
#   node. 
#****
proc spawnShellExec {} {
    upvar 0 ::cf::[set ::curcfg]::eid eid
    global gui_unix

    set node [lindex [.panwin.f1.c gettags {node && current}] 1]
    if { $node == "" } {
	set node [lindex [.panwin.f1.c gettags {nodelabel && current}] 1]
	if { $node == "" } {
	    return
	}
    }
    if { [[typemodel $node].virtlayer] != "VIMAGE" } {
	nodeConfigGUI .panwin.f1.c $node
    } else {
	set cmd [lindex [existingShells [[typemodel $node].shellcmds] $node] 0]
	if { $cmd == "" } {
	    return
	}
	spawnShell $node $cmd 
    }
}

#****f* editor.tcl/startethereal
# NAME
#   startethereal -- start ethereal
# SYNOPSIS
#   startethereal $node $iface
# FUNCTION
#   This procedure starts ethereal for the specified node
#   and the specified interface.
# INPUTS
#   * node -- node id
#   * iface -- interface name
#****
proc startethereal { node iface } {
    upvar 0 ::cf::[set ::curcfg]::eid eid

    set interface "$iface@$eid\.$node"
    nexec jexec $eid.$node tcpdump -s 0 -U -w - -i $iface 2>/dev/null | wireshark -o "gui.window_title:$iface@[getNodeName $node] ($eid)" -k -i - &
}

#****f* editor.tcl/button1
# NAME
#   button1 -- button1 clicked
# SYNOPSIS
#   button1 $c $x $y $button
# FUNCTION
#   This procedure is called when a left mouse button is 
#   clicked on the canvas. This procedure selects a new
#   node or creates a new node, depending on the selected 
#   tool.
# INPUTS
#   * c -- tk canvas
#   * x -- x coordinate
#   * y -- y coordinate
#   * button -- the keyboard button that is pressed.
#****
proc button1 { c x y button } {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global activetool newlink curobj changed def_router_model
    global router pc host lanswitch frswitch rj45 hub
    global oval rectangle text freeform newtext
    global lastX lastY
    global background selectbox
    global defLinkColor defLinkWidth
    global resizemode resizeobj

    set x [$c canvasx $x]
    set y [$c canvasy $y]

    set lastX $x
    set lastY $y

    set curobj [$c find withtag current]
    set curtype [lindex [$c gettags current] 0]
    if { $curtype == "node" || $curtype == "oval" ||
	 $curtype == "rectangle" || $curtype == "text" ||
	 $curtype == "freeform" || ( $curtype == "nodelabel" &&
	 [nodeType [lindex [$c gettags $curobj] 1]] == "pseudo") } {
	set node [lindex [$c gettags current] 1]
	set wasselected \
	    [expr {[lsearch [$c find withtag "selected"] \
	    [$c find withtag "(node || text || freeform || rectangle || oval) && $node"]] > -1}]
	if { $button == "ctrl" } {
	    if { $wasselected } {
		$c dtag $node selected
		$c delete -withtags "selectmark && $node"
	    }
	} elseif { !$wasselected } {
	    foreach node_type { "node" "text" "oval" "rectangle" "freeform"} {
		$c dtag $node_type selected
	    }
	    $c delete -withtags selectmark
	}
	if { $activetool == "select" && !$wasselected} {
	    selectNode $c $curobj
	}
    } elseif { $curtype == "selectmark" } {

	set t1 [$c gettags current]
	set o1 [lindex $t1 1]
	set type1 [nodeType $o1]
    
	if {$type1== "oval" || $type1== "rectangle"} { 
	    set resizeobj $o1
	    set bbox1 [$c bbox $o1]
	    set x1 [lindex $bbox1 0]
	    set y1 [lindex $bbox1 1]
	    set x2 [lindex $bbox1 2]
	    set y2 [lindex $bbox1 3]
	    set l 0 ;# left
	    set r 0 ;# right
	    set u 0 ;# up
	    set d 0 ;# down

	    if { $x < [expr $x1+($x2-$x1)/8.0]} { set l 1 }
	    if { $x > [expr $x2-($x2-$x1)/8.0]} { set r 1 }
	    if { $y < [expr $y1+($y2-$y1)/8.0]} { set u 1 }
	    if { $y > [expr $y2-($y2-$y1)/8.0]} { set d 1 }

	    if {$l==1} {
		if {$u==1} { 
		    set resizemode lu
		} elseif {$d==1} { 
		    set resizemode ld
		} else { 
		    set resizemode l
		} 
	    } elseif {$r==1} {
		if {$u==1} { 
		    set resizemode ru
		} elseif {$d==1} { 
		    set resizemode rd
		} else { 
		    set resizemode r
		} 
	    } elseif {$u==1} { 
		set resizemode u
	    } elseif {$d==1} {
		set resizemode d
	    } else {
		set resizemode false
	    }
	}

    } elseif { $button != "ctrl" || $activetool != "select" } {
	foreach node_type { "node" "text" "oval" "rectangle" "freeform"} {
	    $c dtag $node_type selected
	}
	$c delete -withtags selectmark
    }
    #determine whether we can create nodes on the current object
    set isObjectDrawable 0
    foreach type {background grid rectangle oval freeform text} {
	if { $type in [.panwin.f1.c gettags $curobj] } {
	    set isObjectDrawable 1
	    break
	}
    }
    if { $isObjectDrawable } {
	if { $activetool ni {select link oval rectangle text freeform} } {
	    # adding a new node
	    set node [newNode $activetool]
	    setNodeCanvas $node $curcanvas
	    setNodeCoords $node "[expr {$x / $zoom}] [expr {$y / $zoom}]"
	    # To calculate label distance we take into account the normal icon
	    # height
	    set tempimg [image create photo -file [$activetool.icon normal]]
	    set dy [expr [image height $tempimg]/2 + 16]
	    image delete $tempimg
	    setNodeLabelCoords $node "[expr {$x / $zoom}] \
		[expr {$y / $zoom + $dy}]"
	    drawNode $node
	    selectNode $c [$c find withtag "node && $node"]
	    set changed 1
	} elseif { $activetool == "select" \
	    && $curtype != "node" && $curtype != "nodelabel"} {
	    $c config -cursor cross
	    set lastX $x
	    set lastY $y
	    if {$selectbox != ""} {
		# We actually shouldn't get here!
		$c delete $selectbox
		set selectbox ""
	    }
	} elseif { $activetool == "oval" || $activetool == "rectangle" } {
	    $c config -cursor cross
	    set lastX $x
	    set lastY $y
	} elseif { $activetool == "text" } {
	    $c config -cursor xterm
	    set lastX $x
	    set lastY $y
	    set newtext [$c create text $lastX $lastY -text "" \
		-anchor w -justify left -tags "newtext"]
	}
    } else {
	if {$curtype in {node nodelabel text oval rectangle freeform}} {
	    $c config -cursor fleur
	}
	if {$activetool == "link" && $curtype == "node"} {
	    $c config -cursor cross
	    set lastX [lindex [$c coords $curobj] 0]
	    set lastY [lindex [$c coords $curobj] 1]
	    set newlink [$c create line $lastX $lastY $x $y \
		-fill $defLinkColor -width $defLinkWidth \
		-tags "link"]
	}
    }

    raiseAll $c
}

#****f* editor.tcl/button1-motion
# NAME
#   button1-motion -- button1 moved
# SYNOPSIS
#   button1-motion $c $x $y 
# FUNCTION
#   This procedure is called when a left mouse button is 
#   pressed and the mouse is moved around the canvas. 
#   This procedure creates new select box, moves the 
#   selected nodes or draws a new link.
# INPUTS
#   * c -- tk canvas
#   * x -- x coordinate
#   * y -- y coordinate
#****
proc button1-motion { c x y } {
    global activetool newlink changed
    global lastX lastY sizex sizey selectbox background
    global newoval newrect newtext newfree resizemode
    set x [$c canvasx $x]
    set y [$c canvasy $y]
    set curobj [$c find withtag current]
    set curtype [lindex [$c gettags current] 0]
    if {$activetool == "link" && $newlink != ""} {
	#creating a new link
	$c coords $newlink $lastX $lastY $x $y
    } elseif { $activetool == "select" && $curtype == "nodelabel" \
	&& [nodeType [lindex [$c gettags $curobj] 1]] != "pseudo" } {
	$c move $curobj [expr {$x - $lastX}] [expr {$y - $lastY}]
	set changed 1
	set lastX $x
	set lastY $y
    } elseif { $activetool == "select" && 
	( $curobj == $selectbox || $curtype == "background" ||
	$curtype == "grid" || ($curobj ni [$c find withtag "selected"] &&
	$curtype != "selectmark")) } {
	#forming the selectbox and resizing
	if {$selectbox == ""} {
	    set err [catch {
		set selectbox [$c create line \
		    $lastX $lastY $x $lastY $x $y $lastX $y $lastX $lastY \
		    -dash {10 4} -fill black -width 1 -tags "selectbox"]
		} error]
	    if { $err != 0 } {
		return
	    }
	    $c raise $selectbox "background || link || linklabel || interface"
	} else {
	    set err [catch {
		$c coords $selectbox \
		    $lastX $lastY $x $lastY $x $y $lastX $y $lastX $lastY
		} error]
	    if { $err != 0 } {
		return
	    }
	}
    # actually we should check if curobj==bkgImage
    } elseif { $activetool == "oval" && ( $curobj == $newoval \
	|| $curobj == $background || $curtype == "background" \
	|| $curtype == "grid")} {
	# Draw a new oval
	if {$newoval == ""} {
	    set newoval [$c create oval $lastX $lastY $x $y \
			-outline blue \
			-dash {10 4} -width 1 -tags "newoval"]
	    $c raise $newoval "background || link || linklabel || interface"
	} else {
	    $c coords $newoval \
		$lastX $lastY $x $y
	}
    } elseif { $activetool == "rectangle" && ( $curobj == $newrect \
	|| $curobj == $background || $curtype == "background" \
	|| $curtype == "oval" || $curtype == "grid")} {
	# Draw a new rectangle
	if {$newrect == ""} {
	    set newrect [$c create rectangle $lastX $lastY $x $y \
		-outline blue \
		-dash {10 4} -width 1 -tags "newrect"]
	    $c raise $newrect "oval || background || link || linklabel || interface"
	} else {
	    $c coords $newrect $lastX $lastY $x $y
	}
    } elseif { $activetool == "freeform" && ( $curobj == $newfree \
	|| $curobj == $background || $curtype == "background" \
	|| $curtype == "oval" || $curtype == "rectangle"  \
	|| $curtype == "grid")} {
	# Draw a new freeform
	if {$newfree == ""} {
	    set newfree [$c create line $lastX $lastY $x $y \
		-fill blue -width 2 -tags "newfree"]
	    $c raise $newfree "oval || rectangle || background || link || linklabel || interface"
	} else {
	    xpos $newfree $x $y 2 blue
	}		
    } elseif { $curtype == "selectmark" } {
	foreach o [$c find withtag "selected"] { 
	    set node [lindex [$c gettags $o] 1]
	    set tagovi [$c gettags $o]
	    set koord [getNodeCoords $node]

	    set oldX1 [lindex $koord 0]
	    set oldY1 [lindex $koord 1]
	    set oldX2 [lindex $koord 2]
	    set oldY2 [lindex $koord 3]
	    switch -exact -- $resizemode {
		lu {
		    set oldX1 $x
		    set oldY1 $y
		}
		ld {
		    set oldX1 $x
		    set oldY2 $y
		}
		l {
		    set oldX1 $x
		}
		ru {
		    set oldX2 $x
		    set oldY1 $y
		}
		rd {
		    set oldX2 $x
		    set oldY2 $y
		}
		r {
		    set oldX2 $x
		}
		u {
		    set oldY1 $y
		}
		d {
		    set oldY2 $y
		}
	    }
	    if {$selectbox == ""} {
		set err [catch {
		    set selectbox [$c create line \
			$oldX1 $oldY1 $oldX2 $oldY1 $oldX2 $oldY2 $oldX1 $oldY2 $oldX1 $oldY1 \
			-dash {10 4} -fill black -width 1 -tags "selectbox"]
		    } error]
		if { $err != 0 } {
		    return
		}
		$c raise $selectbox "background || link || linklabel || interface"
	    } else {
		set err [catch {
		    $c coords $selectbox \
			$oldX1 $oldY1 $oldX2 $oldY1 $oldX2 $oldY2 $oldX1 $oldY2 $oldX1 $oldY1
		    } error]
		if { $err != 0 } {
		    return
		}
	    }
	}
    } else {
	foreach img [$c find withtag "selected"] {
	    $c move $img [expr {$x - $lastX}] [expr {$y - $lastY}]

	    set node [lindex [$c gettags $img] 1]

	    foreach elem { "selectmark" "nodelabel" "link"} {
		set obj [$c find withtag "$elem && $node"]
		$c move $obj [expr {$x - $lastX}] [expr {$y - $lastY}]
		if { $elem == "link" } {
		    $c addtag need_redraw withtag "link && $node"
		}
	    }
	}
	foreach link [$c find withtag "link && need_redraw"] {
	    redrawLink [lindex [$c gettags $link] 1]
	}
	$c dtag link need_redraw
	set changed 1
	set lastX $x
	set lastY $y
    }
}

#****f* editor.tcl/xpos
# NAME
#   xpos -- interpolates data for freeform annotations
# SYNOPSIS
#   xpos $tempfree $x $y $width $color
# FUNCTION
#   This procedure is used to interpolate data for freeform annotations to
#   reduce the amount of fixed points that need to be saved for freeform
#   annotations.
# INPUTS
#   * tempfree -- temporary freefrom when drawing
#   * x -- endpoint x coordinate
#   * y -- endpoint y coordinate
#   * width -- freeform width
#   * color -- freeform color
#****
proc xpos { tempfree x y width color } {

    set all_dots [.panwin.f1.c coords $tempfree]
    set len [llength $all_dots]

    # Remove dots one very close to another
    set d 1.5
    set i [expr $len - 20]
    if {$i < 2} {
	set i 2
    }
    for {} {$i < $len} {incr i 2} {
	set a_x [lindex $all_dots [expr $i - 2]]
	set a_y [lindex $all_dots [expr $i - 1]]
	set b_x [lindex $all_dots $i]
	set b_y [lindex $all_dots [expr $i + 1]]
	if {[expr abs($a_x - $b_x)] < $d && [expr abs($a_y - $b_y)] < $d} {
	    set all_dots [lreplace $all_dots $i [expr $i + 1]]
	    incr len -2
	}
    }

    # Remove dots which can be safely linearly interpolated
    set d 1.5
    set i [expr $len - 20]
    if {$i < 2} {
	set i 2
    }
    for {} {$i < $len} {incr i 2} {
	set a_x [lindex $all_dots [expr $i - 4]]
	set a_y [lindex $all_dots [expr $i - 3]]
	set b_x [lindex $all_dots [expr $i - 2]]
	set b_y [lindex $all_dots [expr $i - 1]]
	set c_x [lindex $all_dots $i]
	set c_y [lindex $all_dots [expr $i + 1]]
	if {[expr abs(($a_x + $c_x) / 2 - $b_x)] < $d &&
	    [expr abs(($a_y + $c_y) / 2 - $b_y)] < $d} {
	    set all_dots [lreplace $all_dots [expr $i - 2] [expr $i - 1]]
	    incr len -2
	}
    }

    .panwin.f1.c coords $tempfree [concat $all_dots $x $y]
    .panwin.f1.c itemconfigure $tempfree -fill $color -width $width -capstyle round
}

#****f* editor.tcl/pseudo.layer
# NAME
#   pseudo.layer -- pseudo layer
# SYNOPSIS
#   set layer [pseudo.layer]
# FUNCTION
#   Returns the layer on which the pseudo node operates
#   i.e. returns no layer.
# RESULT
#   * layer -- returns an empty string
#****
proc pseudo.layer {} {
}

#****f* editor.tcl/newGUILink
# NAME
#   newGUILink -- new GUI link
# SYNOPSIS
#   newGUILink $lnode1 $lnode2
# FUNCTION
#   This procedure is called to create a new link between 
#   nodes lnode1 and lnode2. Nodes can be on the same canvas 
#   or on different canvases. The result of this function
#   is directly visible in GUI.
# INPUTS
#   * lnode1 -- node id of the first node
#   * lnode2 -- node id of the second node
#****
proc newGUILink { lnode1 lnode2 } {
    global changed

    set link [newLink $lnode1 $lnode2]
    if { $link == "" } {
	return
    }
    if { [getNodeCanvas $lnode1] != [getNodeCanvas $lnode2] } {
	set new_nodes [splitLink $link pseudo]
	set orig_nodes [linkPeers $link]
	set new_node1 [lindex $new_nodes 0]
	set new_node2 [lindex $new_nodes 1]
	set orig_node1 [lindex $orig_nodes 0]
	set orig_node2 [lindex $orig_nodes 1]
	set new_link1 [linkByPeers $orig_node1 $new_node1]
	set new_link2 [linkByPeers $orig_node2 $new_node2]
	setNodeMirror $new_node1 $new_node2
	setNodeMirror $new_node2 $new_node1
	setNodeName $new_node1 $orig_node2
	setNodeName $new_node2 $orig_node1
	setLinkMirror $new_link1 $new_link2
	setLinkMirror $new_link2 $new_link1
    }
    redrawAll
    set changed 1
    updateUndoLog
}

#****f* editor.tcl/button1-release
# NAME
#   button1-release -- button1 released
# SYNOPSIS
#   button1-release $c $x $y 
# FUNCTION
#   This procedure is called when a left mouse button is 
#   released. 
#   The result of this function depends on the actions
#   during the button1-motion procedure.
# INPUTS
#   * c -- tk canvas
#   * x -- x coordinate
#   * y -- y coordinate
#****
proc button1-release { c x y } {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    upvar 0 ::cf::[set ::curcfg]::undolevel undolevel
    upvar 0 ::cf::[set ::curcfg]::redolevel redolevel
    upvar 0 ::cf::[set ::curcfg]::undolog undolog
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    global activetool newlink curobj grid
    global changed selectbox
    global lastX lastY sizex sizey
    global autorearrange_enabled
    global resizemode resizeobj

    set redrawNeeded 0

    set outofbounds 0
	
    set x [$c canvasx $x]
    set y [$c canvasy $y]

    $c config -cursor left_ptr
    # if the link tool is active and we are creating a new link
    if {$activetool == "link" && $newlink != ""} {
	$c delete $newlink
	set newlink ""
	set destobj ""
	# find the node that is under the cursor
	foreach obj [$c find overlapping $x $y $x $y] {
	    if {[lindex [$c gettags $obj] 0] == "node"} {
		set destobj $obj
		break
	    }
	}
	# if there is an object beneath the cursor and an object was
	# selected by the button1 procedure create a link between nodes
	if {$destobj != "" && $curobj != "" && $destobj != $curobj} {
	    set lnode1 [lindex [$c gettags $curobj] 1]
	    set lnode2 [lindex [$c gettags $destobj] 1]
	    if { [ifcByLogicalPeer $lnode1 $lnode2] == "" } {
		set link [newLink $lnode1 $lnode2]
		if { $link != "" } {
		    drawLink $link
		    redrawLink $link
		    updateLinkLabel $link
		    set changed 1
		}
	    }
	}
    } elseif {$activetool == "rectangle" || $activetool == "oval" \
	|| $activetool == "text" || $activetool =="freeform"} {
	popupAnnotationDialog $c 0 "false" 
    }

    if { $changed == 1 } {
	set regular true
	# selects the node whose label was moved
	if { [lindex [$c gettags $curobj] 0] == "nodelabel" } {
	    set node [lindex [$c gettags $curobj] 1]
	    selectNode $c [$c find withtag "node && $node"]
	}
	set selected {}
	foreach img [$c find withtag "selected"] {
	    set node [lindex [$c gettags $img] 1]
	    lappend selected $node
	    set coords [$c coords $img]
	    set x [expr {[lindex $coords 0] / $zoom}]
	    set y [expr {[lindex $coords 1] / $zoom}]
	    
	    # only nodes are snapped to grid, annotations are not
	    if { $autorearrange_enabled == 0 && \
		[$c find withtag "node && $node"] != "" } {
		set dx [expr {(int($x / $grid + 0.5) * $grid - $x) * $zoom}]
		set dy [expr {(int($y / $grid + 0.5) * $grid - $y) * $zoom}]
		$c move $img $dx $dy
		set coords [$c coords $img]
		set x [expr {[lindex $coords 0] / $zoom}]
		set y [expr {[lindex $coords 1] / $zoom}]
		setNodeCoords $node "$x $y"
		#moving the nodelabel assigned to the moving node
		$c move "nodelabel && $node" $dx $dy
		set coords [$c coords "nodelabel && $node"]
		set x [expr {[lindex $coords 0] / $zoom}]
		set y [expr {[lindex $coords 1] / $zoom}]
		setNodeLabelCoords $node "$x $y"
		if {$x < 0 || $y < 0 || $x > $sizex || $y > $sizey} {
		    set regular false
		}
	    } else {
		set dx 0
		set dy 0
	    }
	    # only nodes (annotations can not) can set the regular flag
#	    if {$x < 0 || $y < 0 || $x > $sizex || $y > $sizey} {
#		set regular false
#	    } 
	    if { [lindex [$c gettags $node] 0] == "oval"} {
		set coordinates [$c coords [lindex [$c gettags $node] 1]]
		set x1 [expr {[lindex $coordinates 0] / $zoom}]
		set y1 [expr {[lindex $coordinates 1] / $zoom}]
		set x2 [expr {[lindex $coordinates 2] / $zoom}]
		set y2 [expr {[lindex $coordinates 3] / $zoom}]
		if {$x1<0} {
		    set x2 [expr {$x2-$x1}]
		    set x1 0
		    set outofbounds 1
		}
		if {$y1<0} {
		    set y2 [expr {$y2-$y1}]
		    set y1 0
		    set outofbounds 1
		}
		if {$x2>$sizex} {
		    set x1 [expr {$x1-($x2-$sizex)}]
		    set x2 $sizex
		    set outofbounds 1
		}
		if {$y2>$sizey} {
		    set y1 [expr {$y1-($y2-$sizey)}]
		    set y2 $sizey
		    set outofbounds 1
		}
		setNodeCoords $node "$x1 $y1 $x2 $y2"
	    }
	    if {[lindex [$c gettags $node] 0] == "rectangle" } {
		set coordinates [$c coords [lindex [$c gettags $node] 1]]
		set x1 [expr {[lindex $coordinates 0] / $zoom}]
		set y1 [expr {[lindex $coordinates 1] / $zoom}]
		set x2 [expr {[lindex $coordinates 6] / $zoom}]
		set y2 [expr {[lindex $coordinates 13] / $zoom}]
		if {$x1<0} {
		    set x2 [expr {$x2-$x1}]
		    set x1 0
		    set outofbounds 1
		}
		if {$y1<0} {
		    set y2 [expr {$y2-$y1}]
		    set y1 0
		    set outofbounds 1
		}
		if {$x2>$sizex} {
		    set x1 [expr {$x1-($x2-$sizex)}]
		    set x2 $sizex
		    set outofbounds 1
		}
		if {$y2>$sizey} {
		    set y1 [expr {$y1-($y2-$sizey)}]
		    set y2 $sizey
		    set outofbounds 1
		}
		setNodeCoords $node "$x1 $y1 $x2 $y2"
	    }
	    if { [lindex [$c gettags $node] 0] == "freeform"} {
		set bbox [$c bbox "selectmark && $node"]
		set x1 [expr {[lindex $bbox 0] / $zoom}]
		set y1 [expr {[lindex $bbox 1] / $zoom}]
		set x2 [expr {[lindex $bbox 2] / $zoom}]
		set y2 [expr {[lindex $bbox 3] / $zoom}]
		set shiftx 0	
		set shifty 0	

		if {$x1<0} {
		    set shiftx -$x1
		    set outofbounds 1
		}
		if {$y1<0} {
		    set shifty -$y1
		    set outofbounds 1
		}
		if {$x2>$sizex} {
		    set shiftx [expr $sizex-$x2]
		    set outofbounds 1
		}
		if {$y2>$sizey} {
		    set shifty [expr $sizey-$y2]
		    set outofbounds 1
		}

		set coordinates [$c coords [lindex [$c gettags $node] 1]]
                set l [expr {[llength $coordinates]-1}]
                set newcoords {}
                set i 0

		while {$i<=$l} {
                    set f1 [expr {[lindex $coords $i] * $zoom}]
                    set g1 [expr {[lindex $coords $i+1] * $zoom}]
                    set xx1 [expr $f1+$shiftx]
                    set yy1 [expr $g1+$shifty]

                    lappend newcoords $xx1 $yy1
                    set i [expr {$i+2}]
                }
                setNodeCoords $node $newcoords
	    }
	    if { [lindex [$c gettags $node] 0] == "text"} {
		set bbox [$c bbox "selectmark && $node"]
		set coordinates [$c coords [lindex [$c gettags $node] 1]]
		set x1 [expr [lindex $coordinates 0]]
		set y1 [expr [lindex $coordinates 1]]
		set width [expr [lindex $bbox 2] - [lindex $bbox 0]]
		set height [expr [lindex $bbox 3] - [lindex $bbox 1]]
		if {[lindex $bbox 0]<0} {
		    set x1 5
		    set outofbounds 1
		}
		if {[lindex $bbox 1]<0} {
		    set y1 [expr $height/2]
		    set outofbounds 1
		}
		if {[lindex $bbox 2]>$sizex} {
		    set x1 [expr $sizex-$width+5]
		    set outofbounds 1
		}
		if {[lindex $bbox 3]>$sizey} {
		    set y1 [expr {$sizey-$height/2}]
		    set outofbounds 1
		}
		setNodeCoords $node "$x1 $y1"
	    }

	    $c move "selectmark && $node" $dx $dy
	    $c addtag need_redraw withtag "link && $node"
	    set changed 1
	} ;# end of: foreach img selected

	if {$outofbounds} {
	    redrawAll
	    if {$activetool == "select" } {
		selectNodes $selected
	    }
	}


	if {$regular == "true"} {
	    foreach link [$c find withtag "link && need_redraw"] {
		redrawLink [lindex [$c gettags $link] 1]
                updateLinkLabel [lindex [$c gettags $link] 1]
	    }
	} else {
	    .panwin.f1.c config -cursor watch
	    loadCfg $undolog($undolevel)
	    redrawAll
	    if {$activetool == "select" } {
		selectNodes $selected
	    }
	    set changed 0
	}
	$c dtag link need_redraw

    # $changed!=1
    } elseif {$activetool == "select" } { 
	if {$selectbox == ""} {
	    set x1 $x
	    set y1 $y
	    set autorearrange_enabled 0
	} else {
	    set coords [$c coords $selectbox]
	    set x [lindex $coords 0]
	    set y [lindex $coords 1]
	    set x1 [lindex $coords 4]
	    set y1 [lindex $coords 5]
	    $c delete $selectbox
	    set selectbox ""
	}

	if { $resizemode == "false" } {
	    set enclosed {}
	    catch {$c find enclosed $x $y $x1 $y1} enc_objs
	    foreach obj $enc_objs {
		set tags [$c gettags $obj]
		if {[lindex $tags 0] == "node" && [lsearch $tags selected] == -1} {
		    lappend enclosed $obj
		}
		if {[lindex $tags 0] == "oval" && [lsearch $tags selected] == -1} {
		    lappend enclosed $obj
		}
		if {[lindex $tags 0] == "rectangle" && [lsearch $tags selected] == -1} {
		    lappend enclosed $obj
		}
		if {[lindex $tags 0] == "text" && [lsearch $tags selected] == -1} {
		    lappend enclosed $obj
		}
		if {[lindex $tags 0] == "freeform" && [lsearch $tags selected] == -1} {
		    lappend enclosed $obj
		}
	    }
	    foreach obj $enclosed {
		selectNode $c $obj
	    }
	} else {
	    setNodeCoords $resizeobj "$x $y $x1 $y1"
	    set redrawNeeded 1
	    set resizemode false
	}
    }

    if { $redrawNeeded } {
	set redrawNeeded 0
	redrawAll
    } else {
	raiseAll $c
    }
    update
    updateUndoLog
}

#****f* editor.tcl/raiseAll
# NAME
#   raiseAll -- raise all
# SYNOPSIS
#   raiseAll $c
# FUNCTION
#   Raises all elements on canvas.
# INPUTS
#   * c -- tk canvas
#****
proc raiseAll { c } {
    $c raise grid background
    $c raise rectangle  "grid || background"
    $c raise oval "rectangle || grid || background"
    $c raise link "oval || rectangle || grid || background"
    $c raise freeform "link ||  oval || rectangle || grid || background"
    $c raise route "freeform || link || oval || rectangle || grid || background"
    $c raise linklabel "route || freeform || link || oval || rectangle || grid || background"
    $c raise interface "linklabel || route || freeform || link || oval || rectangle || grid || background"
    $c raise node "interface || linklabel || route || freeform || link || oval || rectangle || grid || background"
    $c raise nodelabel "node || interface || linklabel || route || freeform || link || oval || rectangle || grid || background"
    $c raise text "nodelabel || node || interface || linklabel || route || freeform || link || oval || rectangle || grid || background"
}

#****f* editor.tcl/nodeEnter
# NAME
#   nodeEnter -- node enter
# SYNOPSIS
#   nodeEnter $c
# FUNCTION
#   This procedure prints the node id, node name and 
#   node model (if exists), as well as all the interfaces
#   of the node in the status line. 
#   Information is presented for the node above which is
#   the mouse pointer.  
# INPUTS
#   * c -- tk canvas
#****
proc nodeEnter { c } {
    global activetool
    
    set node [lindex [$c gettags current] 1]
    set err [catch {nodeType $node} error] 
    if { $err != 0 } {
	return
    }

    set type [nodeType $node]
    set name [getNodeName $node]
    set model [getNodeModel $node]
    if { $model != "" } {
	set line "{$node} $name ($model):"
    } else {
	set line "{$node} $name:"
    }
    if { $type != "rj45" } {
	foreach ifc [ifcList $node] {
	    set line "$line $ifc:[getIfcIPv4addr $node $ifc]"
	}
    }
    .bottom.textbox config -text "$line"

    showCfg $c $node
    showRoute $c $node
}

#****f* editor.tcl/linkEnter
# NAME
#   linkEnter -- link enter
# SYNOPSIS
#   linkEnter $c
# FUNCTION
#   This procedure prints the link id, link bandwidth
#   and link delay in the status line.
#   Information is presented for the link above which is
#   the mouse pointer.  
# INPUTS
#   * c -- tk canvas
#****
proc linkEnter {c} {
    upvar 0 ::cf::[set ::curcfg]::link_list link_list
    global activetool

    set link [lindex [$c gettags current] 1]
    if { [lsearch $link_list $link] == -1 } {
	return
    }
    set line "$link: [getLinkBandwidthString $link] [getLinkDelayString $link]"
    .bottom.textbox config -text "$line"
}

#****f* editor.tcl/anyLeave
# NAME
#   anyLeave
# SYNOPSIS
#   anyLeave $c
# FUNCTION
#   This procedure clears the status line.
# INPUTS
#   * c -- tk canvas
#****
proc anyLeave {c} {
    global activetool

    .bottom.textbox config -text ""
    
    $c delete -withtag showCfgPopup
    $c delete -withtag route
}

#****f* editor.tcl/checkIntRange 
# NAME
#   checkIntRange -- check integer range
# SYNOPSIS
#   set check [checkIntRange $str $low $high]
# FUNCTION
#   This procedure checks the input string to see if it is 
#   an integer between the low and high value.
# INPUTS
#   str -- string to check
#   low -- the bottom value
#   high -- the top value
# RESULT
#   * check -- set to 1 if the str is string between low and high
#   value, 0 otherwise.
#****
proc checkIntRange { str low high } {
    if { $str == "" } {
	return 1
    }
    set str [string trimleft $str 0]
    if { $str == "" } {
	set str 0
    }
    if { ![string is integer $str] } {
	return 0
    }
    if { $str < $low || $str > $high } {
	return 0
    }
    return 1
}

#****f* editor.tcl/focusAndFlash 
# NAME
#   focusAndFlash -- focus and flash
# SYNOPSIS
#   focusAndFlash $W $count
# FUNCTION
#   This procedure sets the focus on the bad entry field
#   and on this field it provides an effect of flashing 
#   for approximately 1 second.
# INPUTS
#   * W -- textbox field that caused the bad entry
#   * count -- the parameter that causes flashes.
#   It can be left blank.
#****
proc focusAndFlash {W {count 9}} {
    global badentry

    set fg black
    set bg white

    if { $badentry == -1 } {
	return
    } else {
	set badentry 1
    }
    focus -force $W
    if {$count<1} {
	$W configure -foreground $fg -background $bg
	set badentry 0
    } else {
	if {$count%2} {
	    $W configure -foreground $bg -background $fg
	} else {
	    $W configure -foreground $fg -background $bg
	}
	after 200 [list focusAndFlash $W [expr {$count - 1}]]
    }
}

#****f* editor.tcl/printCanvas
# NAME
#   printCanvas -- print canvas
# SYNOPSIS
#   printCanvas $w
# FUNCTION
#   This procedure is called when the print button in
#   print dialog box is pressed. 
# INPUTS
#   * w -- print dialog widget
#****
proc printCanvas { w } {
    global sizex sizey

    set prncmd [$w.printframe.e1 get]
    destroy $w
    set p [open "|$prncmd" WRONLY]
    puts $p [.panwin.f1.c postscript -height $sizey -width $sizex -x 0 -y 0 -rotate yes -pageheight 297m -pagewidth 210m]
    close $p
}

#****f* editor.tcl/printCanvasToFile
# NAME
#   printCanvasToFile -- print canvas to file
# SYNOPSIS
#   printCanvasToFile $w $entry
# FUNCTION
#   This procedure is called when the print to file
#   button in print to file dialog box is pressed. 
# INPUTS
#   * w -- print to file dialog widget
#   * entry -- file name
#****
proc printCanvasToFile { w entry } {    
    upvar 0 ::cf::[set ::curcfg]::canvas_list canvas_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global printFileType
    
    if { [string match -nocase *.* [$entry get]] != 1} {
	set box "[$entry get]\.$printFileType"
	$entry delete 0 end
	$entry insert 0 $box
    }
    
    set temp [$entry get]
    if { $temp == "" || [string match -nocase *.$printFileType $temp] != 1} {
	return
    }
    
    set start_canvas $curcanvas
    if { $printFileType == "ps" } {
	set psname [$entry get]
    } else {
	set pdfname [$entry get]
	set name [string range $pdfname 0 end-4]
	set psname "$name.ps"
    }

    foreach canvas $canvas_list {
	set p [open "$psname" a+]
	set curcanvas $canvas
	switchCanvas none
	set sizex [expr {[lindex [getCanvasSize $curcanvas] 0]*$zoom}]
	set sizey [expr {[lindex [getCanvasSize $curcanvas] 1]*$zoom}]
	puts $p [.panwin.f1.c postscript -height $sizey -width $sizex -x 0 -y 0 -rotate yes -pageheight 297m -pagewidth 210m]
	close $p
    }
    
    if { $printFileType == "pdf" } {
	exec ps2pdf -dPDFSETTINGS=/screen $psname $pdfname
	exec rm $psname
    }
    
    set curcanvas $start_canvas
    switchCanvas none
    destroy $w
}

#****f* editor.tcl/deleteSelection
# NAME
#   deleteSelection -- delete selection
# SYNOPSIS
#   deleteSelection
# FUNCTION
#   By calling this procedure all the selected nodes in imunes will 
#   be deleted.
#****
proc deleteSelection {} {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    global changed
    global background 
    global viewid
    catch {unset viewid}
    .panwin.f1.c config -cursor watch; update

    foreach lnode [selectedNodes] {
	if { $lnode != "" } {
	    removeGUINode $lnode
	}
	if { [isAnnotation $lnode] } {
	    deleteAnnotation $curcanvas [nodeType $lnode] $lnode
	}
	set changed 1
    }
    raiseAll .panwin.f1.c
    updateUndoLog
    .panwin.f1.c config -cursor left_ptr
    .bottom.textbox config -text ""
}

#****f* editor.tcl/align2grid
# NAME
#   align2grid -- align to grid
# SYNOPSIS
#   align2grid
# FUNCTION
#   Aligns all nodes to grid.
#****
proc align2grid {} {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global sizex sizey grid changed

    set node_objects [.panwin.f1.c find withtag node]
    if { [llength $node_objects] == 0 } {
	return
    }

    set step [expr {$grid * 4}]

    for { set x $step } { $x <= [expr {$sizex - $step}] } { incr x $step } {
	for { set y $step } { $y <= [expr {$sizey - $step}] } { incr y $step } {
	    if { [llength $node_objects] == 0 } {
		set changed 1
		updateUndoLog
		redrawAll
		return
	    }
	    set node [lindex [.panwin.f1.c gettags [lindex $node_objects 0]] 1]
	    set node_objects [lreplace $node_objects 0 0]
	    setNodeCoords $node "$x $y"
	    set dy 32
	    if { [lsearch {router hub lanswitch rj45} \
		[nodeType $node]] >= 0 } {
		set dy 24
	    }
	    setNodeLabelCoords $node "$x [expr {$y + $dy}]"
	}
    }
}

#****f* editor.tcl/rearrange
# NAME
#   rearrange -- rearrange
# SYNOPSIS
#   rearrange $mode
# FUNCTION
#   This procedure rearranges the position of nodes in imunes.
#   It can be used to rearrange all the nodes or only the selected
#   nodes. 
# INPUTS
#   * mode -- when set to "selected" only the selected nodes will be
#   rearranged.
#****
proc rearrange { mode } {
    upvar 0 ::cf::[set ::curcfg]::link_list link_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global autorearrange_enabled sizex sizey

    set autorearrange_enabled 1
    .menubar.tools entryconfigure "Auto rearrange all" -state disabled
    .menubar.tools entryconfigure "Auto rearrange selected" -state disabled
    .bottom.mbuf config -text "autorearrange"
    if { $mode == "selected" } {
	set tagmatch "node && selected"
    } else {
	set tagmatch "node"
    }
    set otime [clock clicks -milliseconds]
    while { $autorearrange_enabled } {
	set ntime [clock clicks -milliseconds]
	if { $otime == $ntime } {
	    set dt 0.001
	} else {
	    set dt [expr {($ntime - $otime) * 0.001}]
	    if { $dt > 0.2 } {
		set dt 0.2
	    }
	    set otime $ntime
	}

	set objects [.panwin.f1.c find withtag $tagmatch]
	set peer_objects [.panwin.f1.c find withtag node]
	foreach obj $peer_objects {
	    set node [lindex [.panwin.f1.c gettags $obj] 1]
	    set coords [.panwin.f1.c coords $obj]
	    set x [expr {[lindex $coords 0] / $zoom}]
	    set y [expr {[lindex $coords 1] / $zoom}]
	    set x_t($node) $x
	    set y_t($node) $y

	    if { $x > 0 } {
		set fx [expr {1000 / ($x * $x + 100)}]
	    } else {
		set fx 10
	    }
	    set dx [expr {$sizex - $x}]
	    if { $dx > 0 } {
		set fx [expr {$fx - 1000 / ($dx * $dx + 100)}]
	    } else {
		set fx [expr {$fx - 10}]
	    }

	    if { $y > 0 } {
		set fy [expr {1000 / ($y * $y + 100)}]
	    } else {
		set fy 10
	    }
	    set dy [expr {$sizey - $y}]
	    if { $dy > 0 } {
		set fy [expr {$fy - 1000 / ($dy * $dy + 100)}]
	    } else {
		set fy [expr {$fy - 10}]
	    }
	    set fx_t($node) $fx
	    set fy_t($node) $fy
	}

	foreach obj $objects {
	    set node [lindex [.panwin.f1.c gettags $obj] 1]
	    set i [lsearch -exact $peer_objects $obj]
	    set peer_objects [lreplace $peer_objects $i $i]
	    set x $x_t($node)
	    set y $y_t($node)
	    foreach other_obj $peer_objects {
		set other [lindex [.panwin.f1.c gettags $other_obj] 1]
		set o_x $x_t($other)
		set o_y $y_t($other)
		set dx [expr {$x - $o_x}]
		set dy [expr {$y - $o_y}]
		set d [expr {hypot($dx, $dy)}]
		set d2 [expr {$d * $d}]
		set p_fx [expr {1000.0 * $dx / ($d2 * $d + 100)}]
		set p_fy [expr {1000.0 * $dy / ($d2 * $d + 100)}]
		if {[linkByPeers $node $other] != ""} {
		    set p_fx [expr {$p_fx - $dx * $d2 * .0000000005}]
		    set p_fy [expr {$p_fy - $dy * $d2 * .0000000005}]
		}
		set fx_t($node) [expr {$fx_t($node) + $p_fx}]
		set fy_t($node) [expr {$fy_t($node) + $p_fy}]
		set fx_t($other) [expr {$fx_t($other) - $p_fx}]
		set fy_t($other) [expr {$fy_t($other) - $p_fy}]
	    }

	    foreach link $link_list {
		set nodes [linkPeers $link]
		if { [getNodeCanvas [lindex $nodes 0]] != $curcanvas ||
		    [getNodeCanvas [lindex $nodes 1]] != $curcanvas ||
		    [getLinkMirror $link] != "" } {
		    continue
		}
		set peers [linkPeers $link]
		set coords0 [getNodeCoords [lindex $peers 0]]
		set coords1 [getNodeCoords [lindex $peers 1]]
		set o_x \
		    [expr {([lindex $coords0 0] + [lindex $coords1 0]) * .5}]
		set o_y \
		    [expr {([lindex $coords0 1] + [lindex $coords1 1]) * .5}]
		set dx [expr {$x - $o_x}]
		set dy [expr {$y - $o_y}]
		set d [expr {hypot($dx, $dy)}]
		set d2 [expr {$d * $d}]
		set fx_t($node) \
		    [expr {$fx_t($node) + 500.0 * $dx / ($d2 * $d + 100)}]
		set fy_t($node) \
		    [expr {$fy_t($node) + 500.0 * $dy / ($d2 * $d + 100)}]
	    }
	}

	foreach obj $objects {
	    set node [lindex [.panwin.f1.c gettags $obj] 1]
	    if { [catch "set v_t($node)" v] } {
		set vx 0.0
		set vy 0.0
	    } else {
		set vx [lindex $v_t($node) 0]
		set vy [lindex $v_t($node) 1]
	    }
	    set vx [expr {$vx + 1000.0 * $fx_t($node) * $dt}]
	    set vy [expr {$vy + 1000.0 * $fy_t($node) * $dt}]
	    set dampk [expr {0.5 + ($vx * $vx + $vy * $vy) * 0.00001}]
	    set vx [expr {$vx * exp( - $dampk * $dt)}]
	    set vy [expr {$vy * exp( - $dampk * $dt)}]
	    set dx [expr {$vx * $dt}]
	    set dy [expr {$vy * $dt}]
	    set x [expr {$x_t($node) + $dx}]
	    set y [expr {$y_t($node) + $dy}]
	    set v_t($node) "$vx $vy"

	    setNodeCoords $node "$x $y"
	    set e_dx [expr {$dx * $zoom}]
	    set e_dy [expr {$dy * $zoom}]
	    .panwin.f1.c move $obj $e_dx $e_dy
	    set img [.panwin.f1.c find withtag "selectmark && $node"]
	    .panwin.f1.c move $img $e_dx $e_dy
	    set img [.panwin.f1.c find withtag "nodelabel && $node"]
	    .panwin.f1.c move $img $e_dx $e_dy
	    set x [expr {[lindex [.panwin.f1.c coords $img] 0] / $zoom}]
	    set y [expr {[lindex [.panwin.f1.c coords $img] 1] / $zoom}]
	    setNodeLabelCoords $node "$x $y"
	    .panwin.f1.c addtag need_redraw withtag "link && $node"
	}
	foreach link [.panwin.f1.c find withtag "link && need_redraw"] {
	    redrawLink [lindex [.panwin.f1.c gettags $link] 1]
	}
	.panwin.f1.c dtag link need_redraw
	update
    }
    .menubar.tools entryconfigure "Auto rearrange all" -state normal
    .menubar.tools entryconfigure "Auto rearrange selected" -state normal
    .bottom.mbuf config -text ""
}

#****f* editor.tcl/switchCanvas 
# NAME
#   switchCanvas -- switch canvas
# SYNOPSIS
#   switchCanvas $direction
# FUNCTION
#   This procedure switches the canvas in one of the defined 
#   directions (previous, next, first and last).
# INPUTS
#   * direction -- the direction of switching canvas. Can be: prev -- 
#   previus, next -- next, first -- first, last -- last.
#****
proc switchCanvas { direction } {
    upvar 0 ::cf::[set ::curcfg]::canvas_list canvas_list
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    global sizex sizey

    set i [lsearch $canvas_list $curcanvas]
    switch -exact -- $direction {
    prev {
	incr i -1
	if { $i < 0 } {
	set curcanvas [lindex $canvas_list end]
	} else {
	set curcanvas [lindex $canvas_list $i]
	}
    }
    next {
	incr i
	if { $i >= [llength $canvas_list] } {
	set curcanvas [lindex $canvas_list 0]
	} else {
	set curcanvas [lindex $canvas_list $i]
	}
    }
    first {
	set curcanvas [lindex $canvas_list 0]
    }
    last {
	set curcanvas [lindex $canvas_list end]
    }
    }

    .panwin.f1.hframe.t delete all
    set x 0
    foreach canvas $canvas_list {
    set text [.panwin.f1.hframe.t create text 0 0 \
	-text "[getCanvasName $canvas]" -tags "text $canvas"]
    set ox [lindex [.panwin.f1.hframe.t bbox $text] 2]
    set oy [lindex [.panwin.f1.hframe.t bbox $text] 3]
    set tab [.panwin.f1.hframe.t create polygon $x 0 [expr {$x + 7}] 18 \
	[expr {$x + 2 * $ox + 17}] 18 [expr {$x + 2 * $ox + 24}] 0 $x 0 \
	-fill #d9d9d9 -tags "tab $canvas"]
    set line [.panwin.f1.hframe.t create line 0 0 $x 0 [expr {$x + 7}] 18 \
	[expr {$x + 2 * $ox + 17}] 18 [expr {$x + 2 * $ox + 24}] 0 999 0 \
	-fill #d9d9d9 -width 2 -tags "line $canvas"]
    .panwin.f1.hframe.t coords $text [expr {$x + $ox + 12}] [expr {$oy + 2}]
    .panwin.f1.hframe.t raise $text
    incr x [expr {2 * $ox + 17}]
    }
    incr x 7
    .panwin.f1.hframe.t raise "$curcanvas"
    .panwin.f1.hframe.t itemconfigure "tab && $curcanvas" -fill #808080
    .panwin.f1.hframe.t configure -scrollregion "0 0 $x 18"
    update
    set width [lindex [.panwin.f1.hframe.t configure -width] 4]
    set lborder [lindex [.panwin.f1.hframe.t bbox "tab && $curcanvas"] 0]
    set rborder [lindex [.panwin.f1.hframe.t bbox "tab && $curcanvas"] 2]
    set lmargin [expr {[lindex [.panwin.f1.hframe.t xview] 0] * $x - 1}]
    set rmargin [expr {[lindex [.panwin.f1.hframe.t xview] 1] * $x + 1}]
    if { $lborder < $lmargin } {
	.panwin.f1.hframe.t xview moveto [expr {1.0 * ($lborder - 10) / $x}]
    }
    if { $rborder > $rmargin } {
	.panwin.f1.hframe.t xview moveto [expr {1.0 * ($rborder - $width + 10) / $x}]
    }

    set sizex [lindex [getCanvasSize $curcanvas] 0]
    set sizey [lindex [getCanvasSize $curcanvas] 1]

    redrawAll
}

#****f* editor.tcl/renameCanvasPopup 
# NAME
#   renameCanvasPopup -- rename canvas popup
# SYNOPSIS
#   renameCanvasPopup
# FUNCTION
#   Tk widget for renaming the canvas. 
#****
proc renameCanvasPopup {} {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas

    set w .entry1
    catch {destroy $w}
    toplevel $w -takefocus 1
    wm transient $w .
    wm resizable $w 0 0

    #update
    #grab $w
    wm title $w "Canvas rename"
    wm iconname $w "Canvas rename"

    #dodan glavni frame "renameframe"
    ttk::frame $w.renameframe
    pack $w.renameframe -fill both -expand 1

    ttk::label $w.renameframe.msg -wraplength 5i -justify left -text "Canvas name:"
    pack $w.renameframe.msg -side top

    ttk::frame $w.renameframe.buttons
    pack $w.renameframe.buttons -side bottom -fill x -pady 2m
    ttk::button $w.renameframe.buttons.print -text "Apply" -command "renameCanvasApply $w"
    ttk::button $w.renameframe.buttons.cancel -text "Cancel" -command "destroy $w"
    pack $w.renameframe.buttons.print $w.renameframe.buttons.cancel -side left -expand 1

    bind $w <Key-Escape> "destroy $w"
    bind $w <Key-Return> "renameCanvasApply $w"

    ttk::entry $w.renameframe.e1
    $w.renameframe.e1 insert 0 [getCanvasName $curcanvas]
    pack $w.renameframe.e1 -side top -pady 5 -padx 10 -fill x
}

#****f* editor.tcl/resizeCanvasPopup
# NAME
#   resizeCanvasPopup -- resize canvas popup
# SYNOPSIS
#   resizeCanvasPopup
# FUNCTION
#   Creates a popup dialog box for resizing canvas.
#****
proc resizeCanvasPopup {} {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas

    set w .entry1
    catch {destroy $w}
    toplevel $w -takefocus 1
    wm transient $w .
    wm resizable $w 0 0
    #update
    #grab $w
    wm title $w "Canvas resize"
    wm iconname $w "Canvas resize"

    set minWidth [lindex [getMostDistantNodeCoordinates] 0]
    set minHeight [lindex [getMostDistantNodeCoordinates] 1]

    #dodan glavni frame "resizeframe"
    ttk::frame $w.resizeframe
    pack $w.resizeframe -fill both -expand 1


    ttk::label $w.resizeframe.msg -wraplength 5i -justify left -text "Canvas size:"
    pack $w.resizeframe.msg -side top

    ttk::frame $w.resizeframe.buttons
    pack $w.resizeframe.buttons -side bottom -fill x -pady 2m
    ttk::button $w.resizeframe.buttons.print -text "Apply" -command "resizeCanvasApply $w"
    ttk::button $w.resizeframe.buttons.cancel -text "Cancel" -command "destroy $w"
    pack $w.resizeframe.buttons.print $w.resizeframe.buttons.cancel -side left -expand 1
    bind $w <Key-Escape> "destroy $w"
    bind $w <Key-Return> "resizeCanvasApply $w"

    ttk::frame $w.resizeframe.size
    pack $w.resizeframe.size -side top -fill x -pady 2m
    ttk::spinbox $w.resizeframe.size.x -width 4 \
	-validate focus -invalidcommand "focusAndFlash %W"
    $w.resizeframe.size.x insert 0 [lindex [getCanvasSize $curcanvas] 0]
    $w.resizeframe.size.x configure -from $minWidth -to 4096 -increment 2 \
	-validatecommand "checkIntRange %P $minWidth 4096"
    ttk::label $w.resizeframe.size.label -text "*"
    ttk::spinbox $w.resizeframe.size.y -width 4 \
	-validate focus -invalidcommand "focusAndFlash %W"
    $w.resizeframe.size.y insert 0 [lindex [getCanvasSize $curcanvas] 1]
    $w.resizeframe.size.y configure -from $minHeight -to 4096 -increment 2 \
	-validatecommand "checkIntRange %P $minHeight 4096"

    pack $w.resizeframe.size.x $w.resizeframe.size.label $w.resizeframe.size.y -side left -pady 5 -padx 2 -fill x
}

#****f* editor.tcl/renameCanvasApply
# NAME
#   renameCanvasApply -- rename canvas apply
# SYNOPSIS
#   renameCanvasApply $w 
# FUNCTION
#   This procedure is called by clicking on apply button in rename 
#   canvas popup dialog box. It renames the current canvas.
# INPUTS
#   * w -- tk widget (rename canvas popup dialog box)
#****
proc renameCanvasApply { w } {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    global changed

    set newname [$w.renameframe.e1 get]
    destroy $w
    if { $newname != [getCanvasName $curcanvas] } {
	set changed 1
    }
    setCanvasName $curcanvas $newname
    switchCanvas none
    updateUndoLog
}

#****f* editor.tcl/resizeCanvasApply
# NAME
#   resizeCanvasApply -- resize canvas apply
# SYNOPSIS
#   resizeCanvasApply $w
# FUNCTION
#   This procedure is called by clicking on apply button in resize 
#   canvas popup dialog box. It resizes the current canvas.
# INPUTS
#   * w -- tk widget (resize canvas popup dialog box)
#****
proc resizeCanvasApply { w } {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    global changed
    
    set x [$w.resizeframe.size.x get]
    set y [$w.resizeframe.size.y get]
    set ix [lindex [getMostDistantNodeCoordinates] 0]
    set iy [lindex [getMostDistantNodeCoordinates] 1]
    
    if { [getCanvasBkg $curcanvas] == "" && $ix <= $x && $iy <= $y} {
	destroy $w
	if { "$x $y" != [getCanvasSize $curcanvas] } {
	    set changed 1
	}
	setCanvasSize $curcanvas $x $y
	switchCanvas none
	updateUndoLog
    } 
}

#****f* editor.tcl/animate
# NAME
#   animate -- animate
# SYNOPSIS
#   animate
# FUNCTION
#   This function animates the selectbox. The animation looks 
#   different for edit and exec mode.
#****
proc animate {} {
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode
    global animatephase

    .panwin.f1.c itemconfigure "selectmark || selectbox" -dashoffset $animatephase
    incr animatephase 2
    if { $animatephase == 100 } {
	set animatephase 0
    }

    if { $oper_mode == "edit" } {
	after 250 animate
    } else {
	after 1500 animate
    }
}

#****f* editor.tcl/zoom
# NAME
#   zoom -- zoom
# SYNOPSIS
#   zoom $dir
# FUNCTION
#   Zooms the canvas up or down.
# INPUTS
#   * dir -- zoom direction (up or down)
#****
proc zoom { dir } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global zoom_stops
    # set i [lsearch $stops $zoom]
    set minzoom [lindex $zoom_stops 0]
    set maxzoom [lindex $zoom_stops [expr [llength $zoom_stops] - 1]]
    switch -exact -- $dir {
	"down" {
	    if {$zoom > $maxzoom} {
		set zoom $maxzoom
	    } elseif {$zoom < $minzoom} {
		; # leave it unchanged
	    } else {
		set newzoom $minzoom
		foreach z $zoom_stops {
		    if {$zoom <= $z} {
			break
		    } else {
			set newzoom $z
		    }
		}
		set zoom $newzoom 
	    }
	    redrawAll
	}
	"up" {
	    if {$zoom < $minzoom} {
		set zoom $minzoom
	    } elseif {$zoom > $maxzoom} {
		; # leave it unchanged
	    } else {
		foreach z [lrange $zoom_stops 1 end] {
		    set newzoom $z
		    if {$zoom < $z} {
			break
		    }
		}
		set zoom $newzoom 
	    }
	    redrawAll
	}
	default {
	    if { $i < [expr [llength $zoom_stops] - 1] } {
		set zoom [lindex $zoom_stops [expr $i + 1]]
		redrawAll
	    }
	}
    }
}

#****h* editor.tcl/double1onGrid
# NAME
#  double1onGrid.tcl -- called on Double-1 click on grid (bind command)
# SYNOPSIS
#  double1onGrid $c $x $y
# FUNCTION
#  As grid is layered above annotations this procedure is used to find 
#  annotation object closest to cursor.
# INPUTS
#   * c -- tk canvas
#   * x -- double click x coordinate
#   * y -- double click y coordinate
#****
proc double1onGrid { c x y } {
    set obj [$c find closest $x $y]
    set tags [$c gettags $obj]
    set node [lindex $tags 1]
    if {[lsearch $tags grid] != -1 || [lsearch $tags background] != -1} {
	return
    }
    # Is this really necessary?
    set coords [getNodeCoords $node] 
    set x1 [lindex $coords 0]
    set y1 [lindex $coords 1]
    set x2 [lindex $coords 2]
    set y2 [lindex $coords 3]
    if {$x < $x1 || $x > $x2 || $y < $y1 || $y > $y2} {
	# cursor is not ON the closest object
	return
    } else {
	annotationConfig $c $node
    }
}

#****f* editor.tcl/setZoom
# NAME
#   setZoom -- set zoom
# SYNOPSIS
#   setZoom $x $y
# FUNCTION
#   Creates a dialog to set zoom.
# INPUTS
#   * x -- zoom x coordinate
#   * y -- zoom y coordinate
#****
proc setZoom { x y } {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::zoom zoom

    set w .entry1
    catch {destroy $w}
    toplevel $w -takefocus 1

    if { $x == 0 && $y == 0 } {
	set screen [wm maxsize .]
	set x [expr {[lindex $screen 0] / 2}]
	set y [expr {[lindex $screen 1] / 2}]
    } else {
	set x [expr {$x + 10}]
	set y [expr {$y - 90}]
    }
    wm geometry $w +$x+$y
    wm title $w "Set zoom %"
    wm iconname $w "Set zoom %"

    #dodan glavni frame "setzoom"
    ttk::frame $w.setzoom
    pack $w.setzoom -fill both -expand 1

    update
    grab $w
    ttk::label $w.setzoom.msg -wraplength 5i -justify left -text "Zoom percentage:"
    pack $w.setzoom.msg -side top

    ttk::frame $w.setzoom.buttons
    pack $w.setzoom.buttons -side bottom -fill x -pady 2m
    ttk::button $w.setzoom.buttons.print -text "Apply" -command "setZoomApply $w"
    ttk::button $w.setzoom.buttons.cancel -text "Cancel" -command "destroy $w"
    pack $w.setzoom.buttons.print $w.setzoom.buttons.cancel -side left -expand 1

    bind $w <Key-Escape> "destroy $w"
    bind $w <Key-Return> "setZoomApply $w"

    ttk::entry $w.setzoom.e1
    $w.setzoom.e1 insert 0 [expr {int($zoom * 100)}]
    pack $w.setzoom.e1 -side top -pady 5 -padx 10 -fill x
}

#****f* editor.tcl/setZoomApply
# NAME
#   setZoomApply -- set zoom apply
# SYNOPSIS
#   setZoomApply $w
# FUNCTION
#   This procedure is called by clicking on apply button in set
#   zoom popup dialog box. It zooms to a specific point.
# INPUTS
#   * w -- tk widget (set zoom popup dialog box)
#****
proc setZoomApply { w } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom

    set newzoom [expr [$w.setzoom.e1 get] / 100.0]
    if { $newzoom != $zoom } {
	set zoom $newzoom
	redrawAll
    }
    destroy $w
}

#****f* editor.tcl/selectZoom
# NAME
#   selectZoom -- select zoom
# SYNOPSIS
#   selectZoom $x $y
# FUNCTION
#   Creates a dialog to select zoom.
# INPUTS
#   * x -- zoom x coordinate
#   * y -- zoom y coordinate
#****
proc selectZoom { x y } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom

    global zoom_stops
    
    set values {}
    foreach z $zoom_stops {
	lappend values [expr {int($z*100)}]
    }

    set w .entry1
    catch {destroy $w}
    toplevel $w -takefocus 1

    if { $x == 0 && $y == 0 } {
	set screen [wm maxsize .]
	set x [expr {[lindex $screen 0] / 2}]
	set y [expr {[lindex $screen 1] / 2}]
    } else {
	set x [expr {$x + 10}]
	set y [expr {$y - 90}]
    }
    wm geometry $w +$x+$y
    wm title $w "Select zoom %"
    wm iconname $w "Select zoom %"
 
    #dodan glavni frame "selectzoom"
    ttk::frame $w.selectzoom
    pack $w.selectzoom -fill both -expand 1

    ttk::frame $w.selectzoom.buttons
    pack $w.selectzoom.buttons -side bottom -fill x -pady 2m
    ttk::button $w.selectzoom.buttons.print -text "Apply" -command "selectZoomApply $w"
    ttk::button $w.selectzoom.buttons.cancel -text "Cancel" -command "destroy $w"
    pack $w.selectzoom.buttons.print $w.selectzoom.buttons.cancel -side left -expand 1

    bind $w <Key-Escape> "destroy $w"
    bind $w <Key-Return> "selectZoomApply $w"

    ttk::combobox $w.selectzoom.e1 -values $values
    $w.selectzoom.e1 insert 0 [expr {int($zoom * 100)}]
    pack $w.selectzoom.e1 -side top -pady 5 -padx 10 -fill x

    update
    focus $w.selectzoom.e1
    grab $w
}

#****f* editor.tcl/selectZoomApply
# NAME
#   selectZoomApply -- applies zoom values when they are selected
# SYNOPSIS
#   selectZoomApply $w
# FUNCTION
#   This procedure is called by clicking on apply button in select
#   zoom popup dialog box.
# INPUTS
#   * w -- tk widget (select zoom popup dialog box)
#****
proc selectZoomApply { w } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global hasIM changed

    set tempzoom [$w.selectzoom.e1 get]
    
    if {!$hasIM} {
	global zoom_stops
	set values {}
	foreach z $zoom_stops {
	    lappend values [expr {int($z*100)}]
	}
	if { $tempzoom > 400 || $tempzoom < 10 } {
	    set tempzoom 100
	}
	if { [lsearch $values $tempzoom] == -1 } {
	    set tempzoom [expr int($tempzoom/10)*10]
	}
    }

    set newzoom [ expr $tempzoom / 100.0]
    if { $newzoom != $zoom } {
	set zoom $newzoom
	redrawAll
	set changed 1
	updateUndoLog
    }
    destroy $w
}

#****f* editor.tcl/changeAddressRange
# NAME
#   changeAddressRange -- change address range
# SYNOPSIS
#   changeAddressRange
# FUNCTION
#   Change address range for selected nodes.
#****
proc changeAddressRange {} {
    global changed changeAddrRange control changeAddressRange
    global copypaste_nodes copypaste_list

    set control 0
    set autorenumber 1
    set changeAddrRange 0
    set changeAddressRange 1

    if { $copypaste_nodes } {
	set selected_nodes $copypaste_list
	set copypaste_nodes 0
    } else {
	set selected_nodes [selectedNodes]
    }
    set link_nodes_selected ""
    set connected_link_layer_nodes ""
    set autorenumber_nodes ""

    #spremanje svih selektiranih link_layer cvorova u listu link_nodes_selected
    foreach node [lsort -dictionary $selected_nodes] {
	if { [[typemodel $node].layer] == "LINK" } {
	    lappend link_nodes_selected $node
	}
    }

    #spremanje svih medjusobno povezanih selektiranih link_layer cvorova kao jedan element liste connected_link_layer_nodes
    foreach link_node $link_nodes_selected {
	set lan_nodes [lsort -dictionary [listLANnodes $link_node {}]]
	if { [lsearch $connected_link_layer_nodes $lan_nodes] == -1 } {
	    lappend connected_link_layer_nodes $lan_nodes
	}
    }

    global autorenumbered_ifcs
    set autorenumbered_ifcs ""

    #dodijeljivanje adresa suceljima spojenima na link_layer cvorove
    foreach element $connected_link_layer_nodes {
	set counter 0
	foreach node $element {
	    set autorenumber_nodes ""
	    foreach ifc [ifcList $node] {
		set peer [peerByIfc $node $ifc]
		if { [[typemodel $peer].layer] != "LINK" && [lsearch $selected_nodes $peer] != -1 } {
		    set peer_ifc [ifcByPeer $peer $node]
		    lappend autorenumber_nodes "$peer $peer_ifc"
		}
	    }
	    foreach el $autorenumber_nodes {
		set n [lindex $el 0]
		set i [lindex $el 1]
		if { $counter == 0 } {
		    set changeAddrRange 1
		}
		autoIPv4addr $n $i
		lappend autorenumbered_ifcs "$n $i"
		incr counter
		set changed 1
		set changeAddrRange 0
	    }
	}
    }

    set autorenumber_nodes ""
    set autorenumber_ifcs ""

    #spremanje svih selektiranih cvorova koji nisu povezani s link_layer cvorom u listu autorenumber_nodes
    foreach node $selected_nodes {
	if { [[typemodel $node].layer] != "LINK" } {
	    foreach ifc [ifcList $node] {
		set peer [peerByIfc $node $ifc]
		if { [[typemodel $peer].layer] != "LINK" && [lsearch $selected_nodes $peer] != -1 } {
		    lappend autorenumber_ifcs "$node $ifc"
		    if { [lsearch $autorenumber_nodes $node] == -1 } {
			lappend autorenumber_nodes $node
		    }
		}
	    }
	}
    }

    #brisanje adresa selektiranih cvorova prije nove dodjele tako da ne bi
    #utjecalo na detekciju postojecih podmreza
    foreach el $autorenumber_ifcs {
	set node [lindex $el 0] 
	set ifc [lindex $el 1]
	setIfcIPv4addr $node $ifc ""
    }

    #dodijeljivanje adresa suceljima koja nisu spojena na link_layer cvorove
    foreach el $autorenumber_ifcs {
	set node [lindex $el 0] 
	set ifc [lindex $el 1]
	set peer [peerByIfc $node $ifc]
	if { [lsearch $autorenumber_nodes $node] < [lsearch $autorenumber_nodes $peer] } {
	    set changeAddrRange 1
	}
	autoIPv4addr $node $ifc
	set changed 1
	set changeAddrRange 0
    }

    set autorenumber 0
    set changeAddressRange 0
    
    foreach node $selected_nodes {
	foreach ifc [ifcList $node] {
	    autoIPv4defaultroute $node $ifc
	}
    }
    
    redrawAll
    updateUndoLog
}

#****f* editor.tcl/changeAddressRange6
# NAME
#   changeAddressRange6 -- change address range (ipv6)
# SYNOPSIS
#   changeAddressRange6
# FUNCTION
#   Change IPv6 address range for selected nodes.
#****
proc changeAddressRange6 {} {
    global changed changeAddrRange6 control changeAddressRange6
    global copypaste_nodes copypaste_list

    set control 0
    set autorenumber 1
    set changeAddrRange6 0
    set changeAddressRange6 1

    if { $copypaste_nodes } {
	set selected_nodes $copypaste_list
	set copypaste_nodes 0
    } else {
	set selected_nodes [selectedNodes]
    }
    set link_nodes_selected ""
    set connected_link_layer_nodes ""
    set autorenumber_nodes ""

    #spremanje svih selektiranih link_layer cvorova u listu link_nodes_selected
    foreach node [lsort -dictionary $selected_nodes] {
	if { [[typemodel $node].layer] == "LINK" } {
	    lappend link_nodes_selected $node
	}
    }

    #spremanje svih medjusobno povezanih selektiranih link_layer cvorova kao jedan element liste connected_link_layer_nodes
    foreach link_node $link_nodes_selected {
	set lan_nodes [lsort -dictionary [listLANnodes $link_node {}]]
	if { [lsearch $connected_link_layer_nodes $lan_nodes] == -1 } {
	    lappend connected_link_layer_nodes $lan_nodes
	}
    }

    global autorenumbered_ifcs6
    set autorenumbered_ifcs6 ""

    #dodijeljivanje adresa suceljima spojenima na link_layer cvorove
    foreach element $connected_link_layer_nodes {
	set counter 0
	foreach node $element {
	    set autorenumber_nodes ""
	    foreach ifc [ifcList $node] {
		set peer [peerByIfc $node $ifc]
		if { [[typemodel $peer].layer] != "LINK" && [lsearch $selected_nodes $peer] != -1 } {
		    set peer_ifc [ifcByPeer $peer $node]
		    lappend autorenumber_nodes "$peer $peer_ifc"
		}
	    }
	    foreach el $autorenumber_nodes {
		set n [lindex $el 0]
		set i [lindex $el 1]
		if { $counter == 0 } {
		    set changeAddrRange6 1
		}
		autoIPv6addr $n $i
		lappend autorenumbered_ifcs6 "$n $i"
		incr counter
		set changed 1
		set changeAddrRange6 0
	    }
	}
    }

    set autorenumber_nodes ""
    set autorenumber_ifcs ""

    #spremanje svih selektiranih cvorova koji nisu povezani s link_layer cvorom u listu autorenumber_nodes
    foreach node $selected_nodes {
	if { [[typemodel $node].layer] != "LINK" } {
	    foreach ifc [ifcList $node] {
		set peer [peerByIfc $node $ifc]
		if { [[typemodel $peer].layer] != "LINK" && [lsearch $selected_nodes $peer] != -1 } {
		    lappend autorenumber_ifcs "$node $ifc"
		    if { [lsearch $autorenumber_nodes $node] == -1 } {
			lappend autorenumber_nodes $node
		    }
		}
	    }
	}
    }

    #brisanje adresa selektiranih cvorova prije nove dodjele tako da ne bi
    #utjecalo na detekciju postojecih podmreza
    foreach el $autorenumber_ifcs {
	set node [lindex $el 0] 
	set ifc [lindex $el 1]
	setIfcIPv6addr $node $ifc ""
    }

    #dodijeljivanje adresa suceljima koja nisu spojena na link_layer cvorove
    foreach el $autorenumber_ifcs {
	set node [lindex $el 0] 
	set ifc [lindex $el 1]
	set peer [peerByIfc $node $ifc]
	if { [lsearch $autorenumber_nodes $node] < [lsearch $autorenumber_nodes $peer] } {
	    set changeAddrRange6 1
	}
	autoIPv6addr $node $ifc
	set changed 1
	set changeAddrRange6 0
    }

    set autorenumber 0
    set changeAddressRange6 0
    
    foreach node $selected_nodes {
	foreach ifc [ifcList $node] {
	    autoIPv6defaultroute $node $ifc
	}
    }
    
    redrawAll
    updateUndoLog
}

#****f* editor.tcl/routerDefaultsApply
# NAME
#   routerDefaultsApply-- router defaults apply
# SYNOPSIS
#   routerDefaultsApply $wi
# FUNCTION
#   This procedure is called when the button apply is pressed in 
#   popup router defaults dialog box.
# INPUTS
#   * wi -- widget
#****
proc routerDefaultsApply { wi } {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode
    global changed router_model routerDefaultsModel router_ConfigModel
    global routerRipEnable routerRipngEnable routerOspfEnable routerOspf6Enable
    global rdconfig

    lset rdconfig 0 $routerRipEnable
    lset rdconfig 1 $routerRipngEnable
    lset rdconfig 2 $routerOspfEnable 
    lset rdconfig 3 $routerOspf6Enable	
    set routerDefaultsModel $router_model 	
    set model quagga
    set selected_node_list [selectedNodes]
    set empty {}

    if { $selected_node_list != $empty } {
	foreach node $selected_node_list {
	    if { $oper_mode == "edit" && [nodeType $node] == "router" } {
		setNodeModel $node $router_model
		set router_ConfigModel $router_model
		if { $router_ConfigModel != "static" } {
		    set ripEnable [lindex $rdconfig 0]
		    set ripngEnable [lindex $rdconfig 1]
		    set ospfEnable [lindex $rdconfig 2]
		    set ospf6Enable [lindex $rdconfig 3]
		    setNodeProtocolRip $node $ripEnable
		    setNodeProtocolRipng $node $ripngEnable
		    setNodeProtocolOspfv2 $node $ospfEnable
		    setNodeProtocolOspfv3 $node $ospf6Enable
		} else {
		    $wi.nbook.nf1.protocols.rip configure -state disabled
		    $wi.nbook.nf1.protocols.ripng configure -state disabled
		    $wi.nbook.nf1.protocols.ospf configure -state disabled
		    $wi.nbook.nf1.protocols.ospf6 configure -state disabled
		}
		set changed 1
	    }
	}		
    } else {
	foreach node $node_list {
	    if { $oper_mode == "edit" && [nodeType $node] == "router"} {
		setNodeModel $node $router_model
		set router_ConfigModel $router_model
		if { $router_ConfigModel != "static" } {
		    set ripEnable [lindex $rdconfig 0]
		    set ripngEnable [lindex $rdconfig 1]
		    set ospfEnable [lindex $rdconfig 2]
		    set ospf6Enable [lindex $rdconfig 3]
		    setNodeProtocolRip $node  $ripEnable
		    setNodeProtocolRipng $node $ripngEnable
		    setNodeProtocolOspfv2 $node $ospfEnable
		    setNodeProtocolOspfv3 $node $ospf6Enable
		} else {
		    $wi.nbook.nf1.protocols.rip configure -state disabled
		    $wi.nbook.nf1.protocols.ripng configure -state disabled
		    $wi.nbook.nf1.protocols.ospf configure -state disabled
		    $wi.nbook.nf1.protocols.ospf6 configure -state disabled
		}
		set changed 1
	    }		
	}		
    }

    if { $changed == 1 } {
	redrawAll
	updateUndoLog
    }	
    destroy $wi	
}

#****f* editor.tcl/setCustomIcon
# NAME
#   setCustomIcon -- set custom icon
# SYNOPSIS
#   setCustomIcon $node $iconName
# FUNCTION
#   Sets the custom icon to a node.
# INPUTS
#   * node -- node to change
#   * iconName -- icon name
#****
proc setCustomIcon { node iconName } {
    upvar 0 ::cf::[set ::curcfg]::$node $node
    global $iconName
    
    set i [lsearch [set $node] "customIcon *"]
    if { $i >= 0 } {
	set $node [lreplace [set $node] $i $i "customIcon $iconName"]
    } else {
	set $node [linsert [set $node] end "customIcon $iconName"]
    }
}

#****f* editor.tcl/getCustomIcon
# NAME
#   getCustomIcon -- get custom icon
# SYNOPSIS
#   getCustomIcon $node
# FUNCTION
#   Returns the custom icon from a node.
# INPUTS
#   * node -- node to get the icon from
#****
proc getCustomIcon { node } {
    upvar 0 ::cf::[set ::curcfg]::$node $node

    return [lindex [lsearch -inline [set $node] "customIcon *"] 1]
}

#****f* editor.tcl/removeCustomIcon
# NAME
#   removeCustomIcon -- remove custom icon
# SYNOPSIS
#   removeCustomIcon $node
# FUNCTION
#   Removes the custom icon from a node.
# INPUTS
#   * node -- node to remove the icon from
#****
proc removeCustomIcon { node } {
    upvar 0 ::cf::[set ::curcfg]::$node $node

    set i [lsearch [set $node] "customIcon *"]
    if { $i >= 0 } {
	set $node [lreplace [set $node] $i $i]
    }
}

#****f* editor.tcl/removeIPv4nodes
# NAME
#   removeIPv4nodes -- remove ipv4 nodes
# SYNOPSIS
#   removeIPv4nodes
# FUNCTION
#   Sets all nodes' IPv4 addresses to empty strings.
#****
proc removeIPv4nodes {} {
    global changed
    set nodelist [selectedNodes]

    foreach node $nodelist {
	setStatIPv4routes $node "" 
	foreach ifc [ifcList $node] {
	    setIfcIPv4addr $node $ifc "" 
	}
    }
    redrawAll
    set changed 1
    updateUndoLog
}

#****f* editor.tcl/removeIPv6nodes
# NAME
#   removeIPv6nodes -- remove ipv6 nodes
# SYNOPSIS
#   removeIPv6nodes
# FUNCTION
#   Sets all nodes' IPv6 addresses to empty strings.
#****
proc removeIPv6nodes {} {
    global changed
    set nodelist [selectedNodes]

    foreach node $nodelist {
	setStatIPv6routes $node "" 
	foreach ifc [ifcList $node] {
	    setIfcIPv6addr $node $ifc "" 
	}
    }
    redrawAll
    set changed 1
    updateUndoLog
}

#****f* editor.tcl/fetchNodeConfiguration 
# NAME
#   fetchNodeConfiguration -- fetches current node configuration 
# SYNOPSIS
#   fetchNodeConfiguration
# FUNCTION
#   This procedure is called when the button3.menu.sett->Fetch Node
#   Configurations button is pressed. It is used to update the selected nodes
#   configurations from the running experiment settings.
#****
proc fetchNodeConfiguration {} {
    upvar 0 ::cf::[set ::curcfg]::eid eid
    set ip6Set 0
    set ip4Set 0

    foreach node [selectedNodes] {
	catch {exec jexec $eid.$node ifconfig} full
	set lines [split $full "\n"]

	foreach line $lines {
	    if {[regexp {^([[:alnum:]]+):.*mtu ([^$]+)$} $line \
		 -> ifc mtuvalue]} {
		setIfcMTU $node $ifc $mtuvalue
		set ip6Set 0
		set ip4Set 0
	    } elseif {[regexp {^\tether ([^ ]+)} $line -> macaddr]} {
		setIfcMACaddr $node $ifc $macaddr
	    } elseif {[regexp {^\tinet6 (?!fe80:)([^ ]+) } $line -> ip6addr]} {
		if {$ip6Set == 0} {
		    setIfcIPv6addr $node $ifc $ip6addr
		    set ip6Set 1
		}
	    } elseif {[regexp {^\tinet ([^ ]+) netmask ([^ ]+) } $line \
		 -> ip4addr netmask]} {
		if {$ip4Set == 0} {
		    set length [ip::maskToLength $netmask]
		    setIfcIPv4addr $node $ifc $ip4addr/$length
		    set ip4Set 1
		}
	    }
	}
    }
    redrawAll
}

#****f* editor.tcl/setDefaultIcon
# NAME
#   setDefaultIcon -- set default icon
# SYNOPSIS
#   setDefaultIcon
# FUNCTION
#   Sets all selected nodes icons to default icons.
#****
proc setDefaultIcon {} {
    global changed
    set nodelist [selectedNodes]

    foreach node $nodelist {
	set icon [getCustomIcon $node]
	removeCustomIcon $node
	removeImageReference $icon $node
    }
    redrawAll
    set changed 1
    updateUndoLog
}

#****f* editor.tcl/button3background
# NAME
#   button3background -- button3 background
# SYNOPSIS
#   button3background $c $x $y
# FUNCTION
#   Popup menu for right click on canvas background.
# INPUTS
#   * c -- tk canvas
#   * x -- x coordinate
#   * y -- y coordinate
#****
proc button3background { c x y } {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    upvar 0 ::cf::[set ::curcfg]::canvas_list canvas_list
    global showBkgImage changed

    .button3menu delete 0 end

    #
    # Show canvas background
    #
    .button3menu add checkbutton -label "Show background" \
    -underline 5 -variable showBkgImage \
    -command { redrawAll }
    
    .button3menu add separator
    #
    # Change canvas background
    #
    .button3menu add command -label "Change background" \
	    -command "changeBkgPopup"
	    
    #
    # Remove canvas background
    #
    .button3menu add command -label "Remove background" \
	    -command "removeCanvasBkg $curcanvas;
		      if {\"[getCanvasBkg $curcanvas]\" !=\"\"} {
			  removeImageReference [getCanvasBkg $curcanvas] $curcanvas
		      }
		      redrawAll;
		      set changed 1;
		      updateUndoLog"
	    
    
    .button3menu.canvases delete 0 end
    
    set m .button3menu.canvases

    set mode normal
    if {[llength $canvas_list] == 1 } {
	set mode disabled
    }
    .button3menu add cascade -label "Set background from:" -menu $m -underline 0 -state $mode
    foreach c $canvas_list {
	set canv_name [getCanvasName $c]
	set canv_bkg [getCanvasBkg $c]
	set curcanvsize [getCanvasSize $curcanvas]
	set othercanvsize [getCanvasSize $c]
	if {$curcanvas != $c && $curcanvsize == $othercanvsize} {
	    $m add command -label "$canv_name" \
	    -command "setCanvasBkg $curcanvas $canv_bkg;
		      setImageReference $canv_bkg $curcanvas
		      redrawAll;
		      set changed 1;
		      updateUndoLog"
	}
    }

    #
    # Finally post the popup menu on current pointer position
    #
    set x [winfo pointerx .]
    set y [winfo pointery .]
    tk_popup .button3menu $x $y
}

#****f* editor.tcl/changeIconPopup
# NAME
#   changeIconPopup -- change icon popup
# SYNOPSIS
#   changeIconPopup
# FUNCTION
#   Creates a popup dialog to change an icon.
#****
proc changeIconPopup {} {
    upvar 0 ::cf::[set ::curcfg]::image_list image_list
    global chicondialog alignCanvasBkg iconsrcfile wi
    global ROOTDIR LIBDIR
    
    set chicondialog .chiconDialog
    catch {destroy $chicondialog}
    toplevel $chicondialog
    wm transient $chicondialog .
    wm resizable $chicondialog 0 0
    wm title $chicondialog "Set custom icon"
    wm iconname $chicondialog "Set custom icon"

    set wi [ttk::frame $chicondialog.changebgframe]
    
    ttk::panedwindow $wi.iconconf -orient horizontal
    pack $wi.iconconf -fill both
    
    #left and right pane
    ttk::frame $wi.iconconf.left -relief groove -borderwidth 3
    ttk::frame $wi.iconconf.right -relief groove -borderwidth 3
    
    #right pane definition
    ttk::frame $wi.iconconf.right.spacer -height 20
    pack $wi.iconconf.right.spacer -fill both
    ttk::label $wi.iconconf.right.l -text "Icon preview"
    pack $wi.iconconf.right.l -anchor center
    
    set prevcan [canvas $wi.iconconf.right.pc -bd 0 -relief sunken -highlightthickness 0 \
    		-width 100 -height 100 -background white]
    pack $prevcan -anchor center
    
    ttk::label $wi.iconconf.right.l2 -text "Size:"
    pack $wi.iconconf.right.l2 -anchor center
    
    #left pane definition
    #upper left frame with treeview
    ttk::frame $wi.iconconf.left.up
    pack $wi.iconconf.left.up -anchor w -expand 1 -fill both
    
    ttk::frame $wi.iconconf.left.up.grid
    pack $wi.iconconf.left.up.grid -expand 1 -fill both
    
    set tree $wi.iconconf.left.up.tree
    ttk::treeview $tree -columns "type" -height 5 -selectmode browse \
    	-xscrollcommand "$wi.iconconf.left.up.hscroll set"\
        -yscrollcommand "$wi.iconconf.left.up.vscroll set"
    ttk::scrollbar $wi.iconconf.left.up.hscroll -orient horizontal -command "$wi.iconconf.left.up.tree xview"
    ttk::scrollbar $wi.iconconf.left.up.vscroll -orient vertical -command "$wi.iconconf.left.up.tree yview"
    
    grid $wi.iconconf.left.up.tree $wi.iconconf.left.up.vscroll -in $wi.iconconf.left.up.grid -sticky nsew
    #disabled for now, if the addition of new columns happens it will be useful
    #grid $wi.iconconf.left.up.hscroll -in $wi.iconconf.left.up.grid -sticky nsew
    grid columnconfig $wi.iconconf.left.up.grid 0 -weight 1
    grid rowconfigure $wi.iconconf.left.up.grid 0 -weight 1
    
    $tree heading #0 -text "Image name"
    $tree column #0 -width 100 -minwidth 100 
    $tree heading type -text "Type" 
    $tree column type -width 90 -stretch 0 -minwidth 90
    focus $tree
    
    foreach file [glob -directory $ROOTDIR/$LIBDIR/icons/normal/ *.gif] {
	set filename [lindex [split $file /] end]
	$tree insert {} end -id $file -text $filename -values [list "library icon"] \
	  -tags "$file"
	$tree tag bind $file <1> \
	  "updateIconPreview $prevcan $wi.iconconf.right.l2 $file
	   set iconsrcfile \"$file\""
    }
    
    foreach img $image_list {
	if {$img != "" && [string match "*img*" $img] == 1 && \
	  [getImageType $img] == "customIcon"} {
	    $tree insert {} end -id $img -text $img -values [list "custom icon"] \
	      -tags "$img"
	    $tree tag bind $img <1> \
	      "updateIconPreview $prevcan $wi.iconconf.right.l2 $img
	       set iconsrcfile $img"
	}
    }
    
    foreach file [glob -directory $ROOTDIR/$LIBDIR/icons/normal/ *.gif] {
	$tree tag bind $file <Key-Up> \
	    "if {![string equal {} [$tree prev $file]]} {
		updateIconPreview $prevcan $wi.iconconf.right.l2 [$tree prev $file]
		set iconsrcfile [$tree prev $file]
	     }"
	$tree tag bind $file <Key-Down> \
	    "if {![string equal {} [$tree next $file]]} {
		updateIconPreview $prevcan $wi.iconconf.right.l2 [$tree next $file]
		set iconsrcfile [$tree next $file]
	     }"
    }
    
    set first [lindex [glob -directory $ROOTDIR/$LIBDIR/icons/normal/ *.gif] 0]
    $tree selection set $first
    $tree focus $first
        
    foreach img $image_list {
	if {$img != "" && [string match "*img*" $img] == 1 && \
	  [getImageType $img] == "customIcon"} {
	    $tree tag bind $img <Key-Up> \
		"if {![string equal {} [$tree prev $img]]} {
		    updateIconPreview $prevcan $wi.iconconf.right.l2 [$tree prev $img]
		    set iconsrcfile [$tree prev $img]
		}"
	    $tree tag bind $img <Key-Down> \
		"if {![string equal {} [$tree next $img]]} {
		    updateIconPreview $prevcan $wi.iconconf.right.l2 [$tree next $img]
		    set iconsrcfile [$tree next $img]
		}"
	}
    }

    #center left frame with label
    ttk::frame $wi.iconconf.left.center
    pack $wi.iconconf.left.center -anchor w
    ttk::label $wi.iconconf.left.center.l -text "Custom icon file:"
    
    #down left frame with entry and button
    ttk::frame $wi.iconconf.left.down
    ttk::frame $wi.iconconf.left.down.left
    ttk::frame $wi.iconconf.left.down.right
    pack $wi.iconconf.left.down -fill both -padx 10
    pack $wi.iconconf.left.down.left $wi.iconconf.left.down.right \
	-side left -anchor n -padx 2
    
    
    ttk::entry $wi.iconconf.left.down.left.e -width 25 -textvariable iconFile
    ttk::button $wi.iconconf.left.down.right.b -text "Browse" -width 8 \
	-command {
	    set fType {
		{{All Images} {.gif}  {}}
		{{All Images} {.png}  {}}
		{{Gif Images} {.gif}  {}}
		{{PNG Images} {.png} {}}
	    }
#		{{All Images} {.jpeg} {}}
#		{{All Images} {.jpg}  {}}
#		{{All Images} {.bmp}  {}}
#		{{All Images} {.tiff} {}}
#		{{All Images} {.ico}  {}}
#		{{Jpeg Images} {.jpg} {}}
#		{{Jpeg Images} {.jpeg} {}}
#		{{Bitmap Images} {.bmp} {}}
#		{{Tiff Images} {.tiff} {}}
#		{{Icon Images} {.ico} {}}
	    global canvasBkgMode wi
	    set chicondialog .chiconDialog
	    set prevcan $wi.iconconf.right.pc
	    set imgsize $wi.iconconf.right.l2
	    set iconsrcfile [tk_getOpenFile -parent $chicondialog -filetypes $fType]
	    $wi.iconconf.left.down.left.e delete 0 end
	    $wi.iconconf.left.down.left.e insert 0 "$iconsrcfile"
	    if {$iconsrcfile != ""} {
		image create photo iconprev -file $iconsrcfile
		set image_h [image height iconprev]
		set image_w [image width iconprev]
		image delete iconprev
		if { $image_h > 100 || $image_w > 100 } {
		    set iconsrcfile ""
		    $wi.iconconf.left.down.left.e delete 0 end
		    $wi.iconconf.left.down.left.e insert 0 "$iconsrcfile"
		    $imgsize configure -text "Size:"
		    tk_dialog .dialog1 "IMUNES error" \
		    "Error: Icon dimensions can't be bigger than 100x100. This image is $image_w*$image_h." \
		    info 0 Dismiss
		    return
		}
		updateIconPreview $prevcan $imgsize $iconsrcfile
	    }
    }
    
    if {$iconsrcfile != ""} {
	set prevcan $wi.iconconf.right.pc
	set imgsize $wi.iconconf.right.l2
	updateIconPreview $prevcan $imgsize $iconsrcfile
    }    
        
    #packing left side
    pack $wi.iconconf.left.center.l -anchor w -pady 2
    pack $wi.iconconf.left.down.left.e -pady 6
    pack $wi.iconconf.left.down.right.b -pady 4
    pack $wi -fill both
    
    
    #adding panes to paned window
    $wi.iconconf add $wi.iconconf.left
    $wi.iconconf add $wi.iconconf.right  

    #lower frame that contains buttons
    ttk::frame $wi.buttons
    pack $wi.buttons -side bottom -fill x -pady 2m
    ttk::button $wi.buttons.apply -text "Apply" -command {
	    popupIconApply $chicondialog $iconsrcfile
    }
    ttk::button $wi.buttons.cancel -text "Cancel" -command "destroy $chicondialog"
    ttk::button $wi.buttons.remove -text "Remove custom icon" -command \
	 "destroy $chicondialog; setDefaultIcon"
    pack $wi.buttons.remove $wi.buttons.cancel $wi.buttons.apply -side right -expand 1
    
    bind $chicondialog <Key-Return> {popupIconApply $chicondialog $iconsrcfile}
    bind $chicondialog <Key-Escape> "destroy $chicondialog"
}

#****f* editor.tcl/updateIconPreview
# NAME
#   updateIconPreview -- update icon preview
# SYNOPSIS
#   updateIconPreview $pc $imgsize $image
# FUNCTION
#   Updates icon preview.
# INPUTS
#   * pc -- selected pc
#   * imgsize -- image size
#   * image -- image file
#****
proc updateIconPreview { pc imgsize image } {
    $pc delete "preview"
    
    if { ![string match -nocase "*.*" $image] } {
	image create photo iconprev -data [getImageData $image]
    } else {
	image create photo iconprev -file $image
    }
    
    set image_h [image height iconprev]
    set image_w [image width iconprev]
    
    $imgsize configure -text "Size: $image_w*$image_h"
    
    $pc create image 50 50 -anchor center -image iconprev -tags "preview"
}

#****f* editor.tcl/popupIconApply
# NAME
#   popupIconApply -- popup icon apply
# SYNOPSIS
#   popupIconApply $dialog $image
# FUNCTION
#   This procedure is called when the button apply is pressed in 
#   change icon popup dialog box.
# INPUTS
#   * dialog -- tk dialog
#   * image -- image file
#****
proc popupIconApply { dialog image } {
    global changed
    if { $image != "" } {
	set nodelist [selectedNodes]
	if { [string match -nocase "*.*" $image] } {
	    set imgname [loadImage $image "" customIcon $image]
	    foreach node $nodelist {
		set icon [getCustomIcon $node]
		if { $icon != "" } {
		    removeImageReference $icon $node
		}
		setCustomIcon $node $imgname
		setImageReference $imgname $node
	    }
	} else {
	    foreach node $nodelist {
		set icon [getCustomIcon $node]
		if { $icon != "" } {
		    removeImageReference $icon $node
		}
		setCustomIcon $node $image
		setImageReference $image $node
	    }
	}
	redrawAll
	set changed 1
	updateUndoLog
    }
    destroy $dialog
}

#****f* editor.tcl/updateCustomIconReferences
# NAME
#   updateCustomIconReferences -- update custom icon references
# SYNOPSIS
#   updateCustomIconReferences
# FUNCTION
#   Updates custom icon references.
#****
proc updateCustomIconReferences {} {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    foreach node $node_list {
	set icon [getCustomIcon $node]
	if { $icon != "" } {
	    setImageReference $icon $node
	}
    }
}

#****f* editor.tcl/updateIconSize
# NAME
#   updateIconSize -- update icon size
# SYNOPSIS
#   updateIconSize
# FUNCTION
#   Updates icon size.
#****
proc updateIconSize {} {
  global all_modules_list
  foreach b $all_modules_list {
    global $b iconSize
    set $b [image create photo -file [$b.icon $iconSize]]
  }
}

#****f* editor.tcl/getMostDistantNodeCoordinates
# NAME
#   getMostDistantNodeCoordinates -- get most distant node coordinates
# SYNOPSIS
#   getMostDistantNodeCoordinates
# FUNCTION
#   Returns the most distant node coordinates.
#****
proc getMostDistantNodeCoordinates {} {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    set x 0
    set y 0
    foreach node $node_list {
	set coords [getNodeCoords $node]
	if {[lindex $coords 0] > $x} {
	    set x [lindex $coords 0] 
	}
	if {[lindex $coords 1] > $y} {
	    set y [lindex $coords 1] 
	}
    }
    set x [expr $x + 25]
    set y [expr $y + 30]
    
    return [list $x $y]
}

#****f* editor.tcl/selectZoomPopupMenu
# NAME
#   selectZoomPopupMenu -- select zoom popup menu
# SYNOPSIS
#   selectZoomPopupMenu $x $y
# FUNCTION
#   Creates select zoom popup menu on (x,y) coordinates
# INPUTS
#   * x -- x coordinate
#   * y -- y coordinate
#****
proc selectZoomPopupMenu { x y } {
    upvar 0 ::cf::[set ::curcfg]::zoom zoom
    global zoom_stops changed
    .button3menu delete 0 end

    set sel_zoom $zoom
    
    foreach z $zoom_stops {
	.button3menu add radiobutton -label [expr {int($z*100)}] \
	  -variable sel_zoom -value $z \
	  -command {
	      upvar 0 ::cf::[set ::curcfg]::zoom zoom
	      set zoom $sel_zoom
	      redrawAll
	      set changed 1
	      updateUndoLog
	  }
    }

    set x [winfo pointerx .]
    set y [winfo pointery .]
    tk_popup .button3menu $x $y
}

#****f* editor.tcl/showCfg
# NAME
#   showCfg -- show configuration
# SYNOPSIS
#   showCfg $c $node
# FUNCTION
#   This procedure shows popup with configuration selected in
#   Show menu of the node above wich is the mouse pointer
# INPUTS
#   * c -- tk canvas
#   * node -- node id
#****
proc showCfg { c node } {
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode
    upvar 0 ::showConfig showCfg
    upvar 0 ::cf::[set ::curcfg]::eid eid
    upvar 0 ::lastObservedNode lastObservedNode

    #Show only if in exec mode
    if { $oper_mode != "exec" } {
    	return
    }
    #Dont draw again if cursor did not move
    if {[winfo pointerxy .] == $lastObservedNode} {
	    return
    }
    set lastObservedNode [winfo pointerxy .]
    #Dont show popup window if 'None' or 'Route' is selected from 
    #the 'Show' menu
    #Also, dont show popup window if there is no node
    if {$showCfg == "None" || $showCfg == "route" || $node == "" } {
    	$c delete -withtag showCfgPopup
	return
    }
    #Dont show popup window if the node virtlayer is different from VIMAGE
    if {[[typemodel $node].virtlayer] != "VIMAGE"} {
    	return
    }
    #Determine node coordinates
    set coords [getNodeCoords $node]
    set x [expr [lindex $coords 0] + 30]
    set y [expr [lindex $coords 1] + 30]
    #Execute command on selected node and save the command output
    set nodeId "$eid.$node"
    set cmd [concat "nexec jexec " $nodeId $showCfg]
    #set label [eval $cmd]
    catch {eval $cmd} label
    set title "$node# $showCfg\n"
    append title $label
    #Call showCfgPopup
    showCfgPopup $c $node $title $x $y    	
}

#****f* editor.tcl/showCfgPopup
# NAME
#   showCfgPopup -- show configure popup
# SYNOPSIS
#   showCfg $c $node $title $x $y
# FUNCTION
#   This procedure shows popup window
# INPUTS
#   * c -- tk canvas
#   * node -- node id
#   * title -- text that is going to be displayed
#   * x -- x coordinate
#   * y -- y coordinate
#****
proc showCfgPopup { c node title x y } {
    #Therecan be shown only one popup at the time
    $c delete -withtag showCfgPopup
    #Show command output
    set popup [$c create text $x $y \
    			-text $title -tag "showCfgPopup" \
			-font "clean" -justify left -anchor nw]
    #Create frame for the command output
    set box [$c bbox $popup]
    set x1 [expr {[lindex $box 0] - 5}]
    set y1 [expr {[lindex $box 1] - 5}]
    set x2 [expr {[lindex $box 2] + 5}]
    set y2 [expr {[lindex $box 3] + 5}]
    $c create rectangle $x1 $y1 $x2 $y2 -fill "#CECECE" -tag "showCfgPopup"
    $c raise $popup
    #If popup goes beyond the canvas borders move it up and/or left
    set width [expr {abs($x2 - $x1)}]
    set height [expr {abs($y2 - $y1)}]
    set maxRight [winfo width $c]
    set maxDown [winfo height $c]
    set change 0
    set newX $x
    set newY $y
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    set sizex [lindex [getCanvasSize $curcanvas] 0]
    set sizey [lindex [getCanvasSize $curcanvas] 1]

    set nodeX [lindex [getNodeCoords $node] 0]
    set nodeY [lindex [getNodeCoords $node] 1]

    set canvasRegion [$c cget -scrollregion]
    set rx1 [lindex $canvasRegion 0]    
    set ry1 [lindex $canvasRegion 1]    
    set rx2 [lindex $canvasRegion 2]    
    set ry2 [lindex $canvasRegion 3]

    set vx1 [expr {round ([lindex [$c xview] 0]*($rx2-$rx1)+$rx1)}]
    set vx2 [expr {round ([lindex [$c xview] 1]*($rx2-$rx1)+$rx1)}]
    set vy1 [expr {round ([lindex [$c yview] 0]*($ry2-$ry1)+$ry1)}]
    set vy2 [expr {round ([lindex [$c yview] 1]*($ry2-$ry1)+$ry1)}]
    
    set vwidth [expr {abs($vx2 - $vx1)}] 
    set vheight [expr {abs($vy2 - $vy1)}] 

    set shift 40

    if {$nodeX > [expr {$vx1 + $vwidth/2 + 10}]} {
	set newX [expr {$vx1+$shift}] 
    } else {
	set newX [expr {$vx2-$width-$shift}] 
    }
    
    if {$nodeY > [expr {$vy1 + $vheight/2 + 10}]} {
	set newY [expr {$vy1+$shift}] 
    } else {
	set newY [expr {$vy2-$height-$shift}] 
    }
    
    if {$nodeX > [expr {$newX-$shift}] && $nodeX < [expr {$newX+$width+$shift}] \
	&& $nodeY > [expr {$newY-$shift}] && $nodeY < \
	[expr {$newY+$height+$shift}] } {
	return
    }

    if {$x2 > $vx2 || $y2 > $vy2} {
	deleteAndShowPopup $c $title $newX $newY
	return
    }
}

#****f* editor.tcl/deleteAndShowPopup
# NAME
#   deleteAndShowPopup -- delete and show popup
# SYNOPSIS
#   deleteAndShowPopup $c $title $x $y
# FUNCTION
#   Deletes configuration popup and creates it again.
# INPUTS
#   * c -- tk canvas
#   * title -- text that is going to be displayed
#   * x -- x coordinate
#   * y -- y coordinate
#****
proc deleteAndShowPopup { c title x y } {
    $c delete -withtag showCfgPopup
    #Show command output
    set popup [$c create text $x $y \
    			-text $title -tag "showCfgPopup" \
			-font "clean" -justify left -anchor nw]
    #Create frame for the command output
    set box [$c bbox $popup]
    set x1 [expr {[lindex $box 0] - 5}]
    set y1 [expr {[lindex $box 1] - 5}]
    set x2 [expr {[lindex $box 2] + 5}]
    set y2 [expr {[lindex $box 3] + 5}]
    $c create rectangle $x1 $y1 $x2 $y2 -fill "#CECECE" -tag "showCfgPopup"
    $c raise $popup
}

#****f* editor.tcl/showRoute
# NAME
#   showRoute -- show route
# SYNOPSIS
#   showRoute $c $node2
# FUNCTION
#   This procedure shows/draws route between two selected nodes.
#   First node is being selected by clicking on it and second is
#   selected by cursor enter. Route is drawn in green color.
# INPUTS
#   * c -- tk canvas
#   * node2 -- second node
#****
proc showRoute { c node2 } {
    global activetool
    upvar 0 ::showConfig showCfg
    upvar 0 ::traceRouteTime traceRouteTime  
    upvar 0 ::cf::[set ::curcfg]::oper_mode oper_mode
    upvar 0 ::cf::[set ::curcfg]::eid eid
    #Route can only be drawn in exec mode
    if {$oper_mode != "exec"} {
	    return
    }
    #Determine selected node
    set selected {}
    foreach obj [.panwin.f1.c find withtag "node && selected"] {
	lappend selected [lindex [.panwin.f1.c gettags $obj] 1]
    }
    #Draw route only if 'Route' option is selected form 'Show' menu
    if { $showCfg == "route"} {
	#Draw route only if one node is selected
    	if {[llength $selected] != 1} {
		if {[llength $selected] != 0} {
	    	    set line "To show route, only one node can be selected."
	    	    .bottom.textbox config -text "$line"
    		}
	} else {
	    set node1 $selected
	    #Draw route only if both nodes work on network layer
	    set type1 [[typemodel $node1].layer]
	    set type2 [[typemodel $node2].layer]
	    if { $node1 != $node2 && $type1 == "NETWORK" && $type2 == "NETWORK"} {
		#User notification
    		set line "Please wait. Route is being calculated."
    		.bottom.textbox config -text "$line"
		after 5 {set t 1}
		vwait t
		#Get second nodes list of interfaces
		set ifcs [lsort -ascii [ifcList $node2]]
		#Make your own traceroute
		set ifc [lindex $ifcs 0]
		set ip [getIfcIPv4addr $node2 $ifc]
		set slashPlace [string first "/" $ip]
		set ipAddr [string range $ip 0 [expr $slashPlace-1]]
		set nodeId "$eid.$node1"
		set hopIP ""
		set hop 0
		set n1 $node1
		set n2 $node1
		set timeExcedeed No
		set cntr 0
		set errDet 0
		while {$hopIP != $ipAddr} {
		    incr hop
		    set cmd [concat "nexec jexec " $nodeId ping -n -c 1 -m $hop -t 1 -o -s 56 $ipAddr]
		    catch {
			eval $cmd 
	    	    } result
		    set adBeg [string first "from" $result]
		    incr adBeg 5
		    set adEnd [string last ":" $result]
		    incr adEnd -1
		    set hopIP [string range $result $adBeg $adEnd]
		    set n1 $n2
		    set n2 [findNode $c $hopIP]
		    if {$n1 == $n2 || $n2 == "" } {
			set errDet 1
			incr hop -1
			incr cntr
			if {$cntr == 3} break
	    	    } else {
			set cntr 0
			set errDet 0
		    }
		    if {$errDet == 0} {
			drawLine $c $n1 $n2

	    	    }
		}
	    	#User notification
    		set line "Route calculation finished."
    		.bottom.textbox config -text "$line"
	    }
	}
    }
}

#****f* editor.tcl/findNode
# NAME
#   findNode -- find node
# SYNOPSIS
#   findNode $c $ipAddr
# FUNCTION
#   Finds the node with the specified IP address.
# INPUTS
#   * c -- tk canvas
#   * ipAddr -- IP address
#****
proc findNode { c ipAddr } {
    #Get list of all network layer nodes
    set i 0
    set nodeList {}
    foreach obj [$c find withtag node] {
    	set node [lindex [$c gettags $obj] 1]
    	set type [[typemodel $node].layer]
    	if { $type == "NETWORK" } {
	    lappend nodeList $node
	    incr i
    	}
    }
    set nodesNum $i
    #Find node with specified IP address
    for {set j 0} {$j < $nodesNum} {incr j} {
    	set node [lindex $nodeList $j]
    	set ifcs [lsort -ascii [ifcList $node]]
    	foreach ifc $ifcs {
    	set ip [getIfcIPv4addr $node $ifc]
    	set slashPlace [string first "/" $ip]
    	set addr [string range $ip 0 [expr $slashPlace-1]]
    	    if {$addr == $ipAddr} {
		return $node
	    }
    	}
    }

}

#****f* editor.tcl/drawLine
# NAME
#   drawLine -- draw line
# SYNOPSIS
#   drawLine $c $node1 $node2
# FUNCTION
#   Draws the line between two nodes.
# INPUTS
#   * c -- tk canvas.
#   * node1 -- first node
#   * node2 -- second node
#****
proc drawLine { c node1 node2 } {
    global activetool
    set xy1 [getNodeCoords $node1]
    set x1 [lindex $xy1 0]
    set y1 [lindex $xy1 1]				
    set xy2 [getNodeCoords $node2]
    set x2 [lindex $xy2 0]
    set y2 [lindex $xy2 1]
    $c create line $x1 $y1 $x2 $y2 -fill green \
    	-width 3 -tags "route"
    raiseAll $c
    after 5 {set t 1}
    vwait t
}

#****f* editor.tcl/topologyElementsTree 
# NAME
#   topologyElementsTree -- topology elements tree
# SYNOPSIS
#   topologyElementsTree
# FUNCTION
#   Creates the tree with all network elements form the topology.
#****
proc topologyElementsTree {} {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::link_list link_list

    global showTree   
    set f .panwin.f2

    if { !$showTree } {
        .panwin forget $f
    }

    if { $showTree } {
	.panwin add $f
	ttk::frame $f.treegrid
	ttk::treeview $f.tree -selectmode browse \
	    -xscrollcommand "$f.hscroll set"\
	    -yscrollcommand "$f.vscroll set"
	ttk::scrollbar $f.hscroll -orient horizontal -command "$f.tree xview"
	ttk::scrollbar $f.vscroll -orient vertical -command "$f.tree yview"
	
	focus $f.tree

	pack $f.treegrid -side right -fill y
	grid $f.tree $f.vscroll -in $f.treegrid -sticky nsew
        grid $f.hscroll -in $f.treegrid -sticky nsew
	grid columnconfig $f.treegrid 0 -weight 1
	grid rowconfigure $f.treegrid 0 -weight 1
	
	#stvaranje columna            
	$f.tree configure -columns { state MAC IPv4 IPv6 canvas }
	$f.tree column #0 -width 200 -stretch 0
	$f.tree column state -width 60 -anchor center -stretch 0
        $f.tree column MAC -width 120 -anchor center -stretch 0
	$f.tree column IPv4 -width 100 -anchor center -stretch 0
	$f.tree column IPv6 -width 100 -anchor center -stretch 0
	$f.tree column canvas -width 60 -anchor center -stretch 0
	$f.tree heading #0 -text "(Expand All)"
	$f.tree heading state -text "State"
        $f.tree heading MAC -text "MAC address"
	$f.tree heading IPv4 -text "IPv4 address"
	$f.tree heading IPv6 -text "IPv6 address"
	$f.tree heading canvas -text "Canvas"


	#punjenje stabla podacima o cvorovima
        global nodetags
	set nodetags ""
	$f.tree insert {} end -id nodes -text "Nodes" -open true -tags nodes
	$f.tree focus nodes
	$f.tree selection set nodes
	foreach node [lsort -dictionary $node_list] {
	    set type [nodeType $node]
	    if { $type != "pseudo" } {
		$f.tree insert nodes end -id $node -text "[getNodeName $node]" -open false -tags $node
		lappend nodetags $node
		$f.tree set $node canvas [getCanvasName [getNodeCanvas $node]]
		foreach ifc [lsort -dictionary [ifcList $node]] {
		    $f.tree insert $node end -id $node$ifc -text "$ifc" -tags $node$ifc
		    $f.tree set $node$ifc state [getIfcOperState $node $ifc]
		    $f.tree set $node$ifc IPv4 [getIfcIPv4addr $node $ifc]
		    $f.tree set $node$ifc IPv6 [getIfcIPv6addr $node $ifc]
                    $f.tree set $node$ifc MAC [getIfcMACaddr $node $ifc]
		}
	    }
	}

	#punjenje stabla podacima o linkovima
	global linktags
	set linktags ""
	$f.tree insert {} end -id links -text "Links" -open false -tags links
	foreach link [lsort -dictionary $link_list] {
	    set n0 [lindex [linkPeers $link] 0]
	    set n1 [lindex [linkPeers $link] 1]
	    set name0 [getNodeName $n0]
	    set name1 [getNodeName $n1]
	    $f.tree insert links end -id $link -text "From $name0 to $name1" -tags $link
	    lappend linktags $link
	}

	global expandtree
	set expandtree 0
	$f.tree heading #0 -command \
	    "expandOrCollapseTree"

	bindEventsToTree

    } else {
	destroy $f.treegrid
	destroy $f.tree $f.vscroll 
	destroy $f.tree $f.hscroll
	destroy $f.buttons
	destroy $f.tree
    } 
}

#****f* editor.tcl/expandOrCollapseTree
# NAME
#   expandOrCollapseTree -- expand or collapse tree
# SYNOPSIS
#   expandOrCollapseTree
# FUNCTION
#   Expands or collapses all tree items.
#****
proc expandOrCollapseTree {} {
    global expandtree
    if { $expandtree == 0 } {
	set expandtree 1
	set f .panwin.f2
	$f.tree heading #0 -text "(Collapse All)"
	$f.tree item nodes -open true
	$f.tree item links -open true
	foreach node [$f.tree children nodes] {
	    $f.tree item $node -open true	
	}
    } else {
	set expandtree 0
	set f .panwin.f2
	$f.tree heading #0 -text "(Expand All)"
	$f.tree item nodes -open false
	$f.tree item links -open false
	foreach node [$f.tree children nodes] {
	    $f.tree item $node -open false	
	}
    }
}

#****f* editor.tcl/bindEventsToTree
# NAME
#   bindEventsToTree -- bind events to tree
# SYNOPSIS
#   bindEventsToTree
# FUNCTION
#   Adds a Tk binding script for the specified 
#   event sequence to the specified tag. 
#****
proc bindEventsToTree {} {
    global nodetags linktags
    set f .panwin.f2
    $f.tree tag bind nodes <Key-Down> \
	"if {[llength $nodetags] != 0} {
	    selectNodeFromTree [lindex $nodetags 0]
	}"

    $f.tree tag bind links <Key-Up> \
	"if {[llength $nodetags] != 0} {
	    selectNodeFromTree [lindex $nodetags end]
	}"

    $f.tree tag bind links <Key-Down> \
	"if {[llength $linktags] != 0} {
	    selectLinkPeersFromTree [lindex $linktags 0]
	}"

    $f.tree tag bind nodes <1> \
	  ".panwin.f1.c dtag node selected; \
	    .panwin.f1.c delete -withtags selectmark"

    $f.tree tag bind links <1> \
	  ".panwin.f1.c dtag node selected; \
	    .panwin.f1.c delete -withtags selectmark"

    foreach n $nodetags {
	set type [nodeType $n]
	global selectedIfc
	$f.tree tag bind $n <1> \
	      "selectNodeFromTree $n"   
	$f.tree tag bind $n <Key-Up> \
	    "if {![string equal {} [$f.tree prev $n]]} {
		selectNodeFromTree [$f.tree prev $n]
	    } else {
		.panwin.f1.c dtag node selected
		.panwin.f1.c delete -withtags selectmark
	    }" 
	$f.tree tag bind $n <Key-Down> \
	    "if {![string equal {} [$f.tree next $n]]} {
		selectNodeFromTree [$f.tree next $n]
	    } else {
		.panwin.f1.c dtag node selected
		.panwin.f1.c delete -withtags selectmark
	    }"           
	$f.tree tag bind $n <Double-1> \
	    "$type.configGUI .panwin.f1.c $n"
	$f.tree tag bind $n <Key-Return> \
	    "$type.configGUI .panwin.f1.c $n"
	foreach ifc [lsort -dictionary [ifcList $n]] {
	    $f.tree tag bind $n$ifc <Double-1> \
		"set selectedIfc $ifc; \
		  $type.configGUI .panwin.f1.c $n; \
		  set selectedIfc \"\""
	    $f.tree tag bind $n$ifc <Key-Return> \
		"set selectedIfc $ifc; \
		  $type.configGUI .panwin.f1.c $n; \
		  set selectedIfc \"\""
	}
    }

    foreach l $linktags {
	$f.tree tag bind $l <1> \
	    "selectLinkPeersFromTree $l"
	$f.tree tag bind $l <Key-Up> \
	    "if {![string equal {} [$f.tree prev $l]]} {
		selectLinkPeersFromTree [$f.tree prev $l]
	    } else {
		.panwin.f1.c dtag node selected
		.panwin.f1.c delete -withtags selectmark
	    }" 
	$f.tree tag bind $l <Key-Down> \
	    "if {![string equal {} [$f.tree next $l]]} {
		selectLinkPeersFromTree [$f.tree next $l]
	    }"
	$f.tree tag bind $l <Double-1> \
	    "link.configGUI .panwin.f1.c $l"
	$f.tree tag bind $l <Key-Return> \
	    "link.configGUI .panwin.f1.c $l"
    }
}

#****f* editor.tcl/selectNodeFromTree
# NAME
#   selectNodeFromTree -- select node from tree
# SYNOPSIS
#   selectNodeFromTree
# FUNCTION
#   Selects icon of the node selected in the topology tree.
#****
proc selectNodeFromTree { n } {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    set canvas [getNodeCanvas $n]
    set curcanvas $canvas
    switchCanvas none
    .panwin.f1.c dtag node selected
    .panwin.f1.c delete -withtags selectmark
    set obj [.panwin.f1.c find withtag "node && $n"]
    selectNode .panwin.f1.c $obj
}

#****f* editor.tcl/selectLinkPeersFromTree
# NAME
#   selectLinkPeersFromTree -- select link peers from tree
# SYNOPSIS
#   selectLinkPeersFromTree
# FUNCTION
#   Selects icons of nodes that are endnodes
#   of the link selected in the topology tree.
#****
proc selectLinkPeersFromTree { l } {
    upvar 0 ::cf::[set ::curcfg]::curcanvas curcanvas
    set n0 [lindex [linkPeers $l] 0]
    set n1 [lindex [linkPeers $l] 1]    
    set canvas [getNodeCanvas $n0]
    set curcanvas $canvas
    switchCanvas none
    .panwin.f1.c dtag node selected
    .panwin.f1.c delete -withtags selectmark
    set obj0 [.panwin.f1.c find withtag "node && $n0"]
    set obj1 [.panwin.f1.c find withtag "node && $n1"]
    selectNode .panwin.f1.c $obj0
    selectNode .panwin.f1.c $obj1
}

#****f* editor.tcl/refreshTopologyTree
# NAME
#   refreshTopologyTree -- refresh topology tree
# SYNOPSIS
#   refreshTopologyTree
# FUNCTION
#   Refreshes the topology tree.
#****
proc refreshTopologyTree {} {
    upvar 0 ::cf::[set ::curcfg]::node_list node_list
    upvar 0 ::cf::[set ::curcfg]::link_list link_list
    global nodetags linktags

    set f .panwin.f2
    set selected [$f.tree selection]

    $f.tree heading #0 -text "(Expand All)"

    $f.tree delete { nodes links }

    set nodetags ""
    $f.tree insert {} end -id nodes -text "Nodes" -open true -tags nodes
    foreach node [lsort -dictionary $node_list] {
	set type [nodeType $node]
	if { $type != "pseudo" } {
	    $f.tree insert nodes end -id $node -text "[getNodeName $node]" -tags $node
	    lappend nodetags $node
	    $f.tree set $node canvas [getCanvasName [getNodeCanvas $node]]
	    foreach ifc [lsort -dictionary [ifcList $node]] {
		    $f.tree insert $node end -id $node$ifc -text "$ifc" -tags $node$ifc
		    $f.tree set $node$ifc state [getIfcOperState $node $ifc]
		    $f.tree set $node$ifc IPv4 [getIfcIPv4addr $node $ifc]
		    $f.tree set $node$ifc IPv6 [getIfcIPv6addr $node $ifc]
                    $f.tree set $node$ifc MAC [getIfcMACaddr $node $ifc]
	    }
	}
    }

    set linktags ""
    $f.tree insert {} end -id links -text "Links" -open false -tags links
    foreach link [lsort -dictionary $link_list] {
	set n0 [lindex [linkPeers $link] 0]
	set n1 [lindex [linkPeers $link] 1]
	set name0 [getNodeName $n0]
	set name1 [getNodeName $n1]
	$f.tree insert links end -id $link -text "From $name0 to $name1" -tags $link
	lappend linktags $link
    }

    if { [$f.tree exists $selected] } {
	$f.tree focus $selected
	$f.tree selection set $selected
    } else {
	$f.tree focus nodes
	$f.tree selection set nodes
    }

    bindEventsToTree
}

#****f* editor.tcl/attachToExperimentPopup
# NAME
#   attachToExperimentPopup -- attach to experiment popup
# SYNOPSIS
#   attachToExperimentPopup
# FUNCTION
#   Creates a popup dialog box to attach to experiment.
#****
proc attachToExperimentPopup {} {
    global selectedExperiment
    set  ateDialog .attachToExperimentDialog
    catch {destroy $ateDialog}
    toplevel $ateDialog
    wm transient $ateDialog .
    wm resizable $ateDialog 0 0
    wm title $ateDialog "Attach to experiment"
    wm iconname $ateDialog "Attach to experiment"

    set wi [ttk::frame $ateDialog.mainframe]
    
    ttk::panedwindow $wi.expChooser -orient horizontal
    pack $wi.expChooser -fill both
    
    #left and right pane
    ttk::frame $wi.expChooser.left -relief groove -borderwidth 3
    pack  $wi.expChooser.left
    ttk::frame $wi.expChooser.right -relief groove -borderwidth 3
    pack  $wi.expChooser.right
    
    #right pane definition
    set prevcan [canvas $wi.expChooser.right.pc -bd 0 -relief sunken -highlightthickness 0 \
    		-width 300 -height 210 -background white]
    pack $prevcan -anchor center
    $prevcan create text 150 105 -text "(Preview)" -tags "preview"
    
    $wi.expChooser add $wi.expChooser.left
    $wi.expChooser add $wi.expChooser.right
    pack $wi

    ttk::frame $wi.expChooser.left.grid
    pack $wi.expChooser.left.grid -expand 1 -fill both
    
    set tree $wi.expChooser.left.tree
    ttk::treeview $tree -columns "type" -height 5 -selectmode browse \
    	-xscrollcommand "$wi.expChooser.left.hscroll set"\
        -yscrollcommand "$wi.expChooser.left.vscroll set"
    ttk::scrollbar $wi.expChooser.left.hscroll -orient horizontal -command "$wi.expChooser.left.tree xview"
    ttk::scrollbar $wi.expChooser.left.vscroll -orient vertical -command "$wi.expChooser.left.tree yview"
    
    grid $wi.expChooser.left.tree $wi.expChooser.left.vscroll -in $wi.expChooser.left.grid -sticky nsew
    #disabled for now, if the addition of new columns happens it will be useful
    #grid $wi.expChooser.left.up.hscroll -in $wi.expChooser.left.up.grid -sticky nsew
    grid columnconfig $wi.expChooser.left.grid 0 -weight 1
    grid rowconfigure $wi.expChooser.left.grid 0 -weight 1
    
    $tree heading #0 -text "Experiment ID"
    $tree column #0 -width 240 -minwidth 100 
    $tree heading type -text "Timestamp" 
    $tree column type -width 200 -stretch 0 -minwidth 90
    focus $tree

    foreach exp [getResumableExperiments] {
	set timestamp [getExperimentTimestampFromFile $exp]
	$tree insert {} end -id $exp -text [list $exp "-" [getExperimentNameFromFile $exp]] -values [list $timestamp] \
	          -tags "$exp"
	$tree tag bind $exp <1> \
	  "updateScreenshotPreview $prevcan /var/run/imunes/$exp/screenshot.png 
	   set selectedExperiment $exp"
    }
    
    foreach exp [getResumableExperiments] {
	$tree tag bind $exp <Key-Up> \
	"if {![string equal {} [$tree prev $exp]]} {
	    updateScreenshotPreview $prevcan /var/run/imunes/[$tree prev $exp]/screenshot.png
	    set selectedExperiment [$tree prev $exp]
	}"
	$tree tag bind $exp <Key-Down> \
	"if {![string equal {} [$tree next $exp]]} {
	    updateScreenshotPreview $prevcan /var/run/imunes/[$tree next $exp]/screenshot.png
	    set selectedExperiment [$tree next $exp]
	}"
    }

    set first [lindex [getResumableExperiments] 0]
    $tree selection set $first
    $tree focus $first
    set selectedExperiment $first

    if {$selectedExperiment != ""} {
	updateScreenshotPreview $prevcan /var/run/imunes/$selectedExperiment/screenshot.png
    }
    
    ttk::frame $wi.buttons
    pack $wi.buttons -side bottom -fill x -pady 2m
    ttk::button $wi.buttons.resume -text "Resume selected experiment" -command {
	global selectedExperiment
	resumeSelectedExperiment $selectedExperiment
	destroy .attachToExperimentDialog 
    }
    ttk::button $wi.buttons.cancel -text "Cancel" -command "destroy $ateDialog"
    pack $wi.buttons.cancel $wi.buttons.resume -side right -expand 1
    
    bind $ateDialog <Key-Return> {resumeSelectedExperiment $selectedExperiment; destroy .attachToExperimentDialog}
    bind $ateDialog <Key-Escape> "destroy $ateDialog"
}

#****f* editor.tcl/updateScreenshotPreview
# NAME
#   updateScreenshotPreview -- update screenshot preview
# SYNOPSIS
#   updateScreenshotPreview $pc $image
# FUNCTION
#   Updates the screenshot preview.
# INPUTS
#   * pc -- selected pc
#   * image -- image file
#****
proc updateScreenshotPreview { pc image } {
    $pc delete "preview"
    if {[file exists $image]} {
	image create photo screenshot -file $image
	$pc create image 150 105 -image screenshot -tags "preview"
    } else {
	$pc create text 150 100 -text "No screenshot available." -tags "preview"
    }
}

#****f* editor.tcl/resumeSelectedExperiment
# NAME
#   resumeSelectedExperiment -- resume selected experiment
# SYNOPSIS
#   resumeSelectedExperiment $exp
# FUNCTION
#   Resumes selected experiment.
# INPUTS
#   * exp -- experiment id
#****
proc resumeSelectedExperiment { exp } {
    upvar 0 ::cf::[set ::curcfg]::eid eid    
    set curr_eid $eid
    if {$curr_eid == $exp} {
	return
    }
    newProject

    upvar 0 ::cf::[set ::curcfg]::currentFile currentFile
    upvar 0 ::cf::[set ::curcfg]::cfgDeployed cfgDeployed
    upvar 0 ::cf::[set ::curcfg]::eid eid
    upvar 0 ::cf::[set ::curcfg]::ngnodemap ngnodemap
    
    set currentFile [getExperimentConfigurationFromFile $exp]
    openFile

    set ngmapFile "/var/run/imunes/$exp/ngnodemap"
    set fileId [open $ngmapFile r]
    array set ngnodemap [gets $fileId]
    close $fileId

    set eid $exp
    set cfgDeployed true
    setOperMode exec
}

#****f* editor.tcl/setActiveTool
# NAME
#   setActiveTool -- set active tool
# SYNOPSIS
#   setActiveTool $tool
# FUNCTION
#   Sets the active tool to $tool and enables/disables
#   the TopoGen submenus.
# INPUTS
#   * tool -- active tool to set
#****
proc setActiveTool { tool } {
    global activetool mf ROOTDIR LIBDIR
    set ungrouped {select link rectangle oval freeform text} 

    if { $activetool in $ungrouped } {
	$mf.left.$activetool state !selected
    } elseif { [$activetool.layer] == "LINK" } {
	$mf.left.link_layer state !selected
    } elseif { [$activetool.layer] == "NETWORK" } {
	$mf.left.net_layer state !selected
    }

    if { $tool in $ungrouped } {
	$mf.left.$tool state selected
    } elseif { [$tool.layer] == "LINK" } {
	set image [image create photo -file [$tool.icon toolbar]]
	set arrowimage [image create photo -file "$ROOTDIR/$LIBDIR/icons/tiny/l2.gif"]
	$image copy $arrowimage -from 29 30 40 40 -to 29 30 40 40 -compositingrule overlay
	$mf.left.link_layer configure -image $image 
	$mf.left.link_layer state selected
    } elseif { [$tool.layer] == "NETWORK" } {
	set image [image create photo -file [$tool.icon toolbar]]
	set arrowimage [image create photo -file "$ROOTDIR/$LIBDIR/icons/tiny/l3.gif"]
	$image copy $arrowimage -from 29 30 40 40 -to 29 30 40 40 -compositingrule overlay
	$mf.left.net_layer configure -image $image 
	$mf.left.net_layer state selected
    }

    set activetool $tool

    if { $tool in {"router" "pc" "host"} } {
	set state normal
    } else {
	set state disabled
    }

    for { set i 0 } { $i <= [.menubar.t_g index last] } { incr i } {
	.menubar.t_g entryconfigure $i -state $state
    }
}
